from collections import defaultdict
import logging
from deltacat import logs
from deltacat.compute.converter.model.convert_input_files import (
    ConvertInputFiles,
    DataFileList,
    DataFileListGroup,
)
from typing import List, Dict, Tuple, Any
from enum import Enum
from pyiceberg.manifest import DataFile

logger = logs.configure_deltacat_logger(logging.getLogger(__name__))


def check_data_files_sequence_number(
    data_files_list: DataFileList,
    equality_delete_files_list: DataFileList,
) -> Tuple[DataFileListGroup, DataFileListGroup]:
    # Sort by file sequence number
    data_files_list.sort(key=lambda file_tuple: file_tuple[0])
    equality_delete_files_list.sort(key=lambda file_tuple: file_tuple[0])
    data_file_delete_applicable = []
    result_eq_files_list = []

    # Loop through each value in equality_delete_file
    for data_file_tuple in data_files_list:

        # Find all values in equality delete file that having a larger sequence number than current data file
        valid_values_eq = []

        # Pointer for equality delete file
        eq_file_pointer = 0
        # Move data_file_pointer to the first value in data_file that is smaller than val_equality
        while (
            eq_file_pointer < len(equality_delete_files_list)
            and equality_delete_files_list[eq_file_pointer][0] > data_file_tuple[0]
        ):
            valid_values_eq.append(equality_delete_files_list[eq_file_pointer])
            eq_file_pointer += 1

        if valid_values_eq:
            # Append the value for both applicable eq files list and applicable data files list
            data_file_delete_applicable.append(data_file_tuple)
            result_eq_files_list.append(valid_values_eq)

    res_data_file_list = []
    res_equality_delete_file_list = []
    merged_file_dict = defaultdict(list)
    for data_file_sublist, eq_delete_sublist in zip(
        data_file_delete_applicable, result_eq_files_list
    ):
        merged_file_dict[tuple(eq_delete_sublist)].append(data_file_sublist)
    for eq_file_list, data_file_list in merged_file_dict.items():
        res_data_file_list.append(list(set(data_file_list)))
        res_equality_delete_file_list.append(list(set(eq_file_list)))

    assert len(res_data_file_list) == len(res_equality_delete_file_list), (
        f"length of applicable data files list: {len(res_data_file_list)} "
        f"should equal to length of equality delete files list:{len(res_equality_delete_file_list)}"
    )

    return res_equality_delete_file_list, res_data_file_list


def construct_iceberg_table_prefix(
    iceberg_warehouse_bucket_name: str, table_name: str, iceberg_namespace: str
) -> str:
    return f"{iceberg_warehouse_bucket_name}/{iceberg_namespace}/{table_name}/data"


def partition_value_record_to_partition_value_string(partition: Any) -> str:
    # Get string representation of partition value out of Record[partition_value]
    partition_value_str = partition.__repr__().split("[", 1)[1].split("]")[0]
    return partition_value_str


def group_all_files_to_each_bucket(
    data_file_dict: Dict[Any, DataFileList],
    equality_delete_dict: Dict[Any, DataFileList],
    pos_delete_dict: Dict[Any, DataFileList],
) -> List[ConvertInputFiles]:
    convert_input_files_for_all_buckets = []
    files_for_each_bucket_for_deletes = defaultdict(tuple)
    if equality_delete_dict:
        for partition_value, equality_delete_file_list in equality_delete_dict.items():
            if partition_value in data_file_dict:
                (
                    result_equality_delete_file,
                    result_data_file,
                ) = check_data_files_sequence_number(
                    data_files_list=data_file_dict[partition_value],
                    equality_delete_files_list=equality_delete_dict[partition_value],
                )
                files_for_each_bucket_for_deletes[partition_value] = (
                    result_data_file,
                    result_equality_delete_file,
                    [],
                )

    for partition_value, all_data_files_for_each_bucket in data_file_dict.items():
        convert_input_file = ConvertInputFiles.of(
            partition_value=partition_value,
            all_data_files_for_dedupe=all_data_files_for_each_bucket,
        )
        if partition_value in files_for_each_bucket_for_deletes:
            convert_input_file.applicable_data_files = (
                files_for_each_bucket_for_deletes[partition_value][0]
            )
            convert_input_file.applicable_equality_delete_files = (
                files_for_each_bucket_for_deletes[partition_value][1]
            )
        convert_input_files_for_all_buckets.append(convert_input_file)
    return convert_input_files_for_all_buckets


def sort_data_files_maintaining_order(data_files: DataFileList) -> DataFileList:
    """
    Sort data files deterministically based on two criterias:
    1. Sequence number: Newly added files will have a higher sequence number
    2. File path: If file sequence is the same, files are guaranteed to be returned in a deterministic order since file path is unique.
    """
    if data_files:
        data_files = sorted(data_files, key=lambda f: (f[0], f[1].file_path))
    return data_files


class SnapshotType(Enum):
    """Enumeration of possible snapshot types."""

    NONE = "none"
    APPEND = "append"
    REPLACE = "replace"
    DELETE = "delete"


def _get_snapshot_action_description(
    snapshot_type: SnapshotType,
    files_to_delete: List[List[DataFile]],
    files_to_add: List[DataFile],
) -> str:
    """Get a human-readable description of the snapshot action."""
    descriptions = {
        SnapshotType.NONE: "No changes needed",
        SnapshotType.APPEND: f"Adding {len(files_to_add)} new files",
        SnapshotType.REPLACE: f"Replacing {sum(len(files) for files in files_to_delete)} files with {len(files_to_add)} new files",
        SnapshotType.DELETE: f"Deleting {sum(len(files) for files in files_to_delete)} files",
    }
    return descriptions[snapshot_type]


def _determine_snapshot_type(
    to_be_deleted_files: List[List[DataFile]], to_be_added_files: List[DataFile]
) -> SnapshotType:
    """
    Determine the snapshot type based on file changes.

    Args:
        to_be_deleted_files: List of files to be deleted
        to_be_added_files: List of files to be added

    Returns:
        SnapshotType indicating what kind of snapshot to commit
    """
    has_files_to_delete = bool(to_be_deleted_files)
    has_files_to_add = bool(to_be_added_files)

    if not has_files_to_delete and not has_files_to_add:
        return SnapshotType.NONE
    elif not has_files_to_delete and has_files_to_add:
        return SnapshotType.APPEND
    elif has_files_to_delete and has_files_to_add:
        return SnapshotType.REPLACE
    else:  # has_files_to_delete and not has_files_to_add
        return SnapshotType.DELETE



import os
import logging

import uuid
import daft
from pyiceberg.catalog import CatalogType

import deltacat as dc

from deltacat import logs
from deltacat import IcebergCatalog
from deltacat.experimental.catalog.iceberg import IcebergCatalogConfig
from env import store_cli_args_in_os_environ

from pyiceberg.schema import (
    Schema,
    NestedField,
    DoubleType,
    StringType,
)
from pyiceberg.partitioning import PartitionSpec, PartitionField
from pyiceberg.transforms import BucketTransform

from deltacat.experimental.storage.iceberg.model import (
    SchemaMapper,
    PartitionSchemeMapper,
)
from deltacat.env import create_ray_runtime_environment

# initialize the driver logger
driver_logger = logs.configure_application_logger(logging.getLogger(__name__))


def run(warehouse="s3://my-bucket/my/key/prefix", **kwargs):
    """
    This is an e2e example that
        1. creates a DeltaCAT Table (backed by an Iceberg Table) in Glue
        2. writes data into the DeltaCAT Table
        3. reads data from the DeltaCAT Table using Daft

    To run the script:
        1. prepare an AWS Account
            1. prepare a S3 location where the data will be written to, which will be used in Step 3.
            2. prepare an IAM Role that has access to the S3 location and Glue
        2. retrieve the IAM Role AWS Credential and cache locally in ~/.aws/credentials
        3. run below command to execute the example
        ```
        make venv && source venv/bin/activate
        python -m deltacat.examples.iceberg.iceberg_bucket_writer --warehouse=s3://<YOUR_S3_LOCATION>
        ```

    """
    # create any runtime environment required to run the example
    runtime_env = create_ray_runtime_environment()

    # Start by initializing DeltaCAT and registering available Catalogs.
    # Ray will be initialized automatically via `ray.init()`.
    # Only the `iceberg` data catalog is provided so it will become the default.
    # If initializing multiple catalogs, use the `default_catalog_name` param
    # to specify which catalog should be the default.

    dc.init(
        catalogs={
            # the name of the DeltaCAT catalog is "iceberg"
            "iceberg": dc.Catalog(
                # Apache Iceberg implementation of deltacat.catalog.interface
                impl=IcebergCatalog,
                # kwargs for pyiceberg.catalog.load_catalog start here...
                # the name of the Iceberg catalog is "example-iceberg-catalog"
                name="example-iceberg-catalog",
                # for additional properties see:
                # https://py.iceberg.apache.org/configuration/
                config=IcebergCatalogConfig(
                    type=CatalogType.GLUE,
                    properties={
                        "warehouse": warehouse,
                        "region_name": "us-east-1",
                    },
                ),
            )
        },
        # pass the runtime environment into ray.init()
        ray_init_args={"runtime_env": runtime_env},
    )

    # define a native Iceberg table schema
    schema = Schema(
        NestedField(field_id=1, name="symbol", field_type=StringType(), required=True),
        NestedField(field_id=2, name="bid", field_type=DoubleType(), required=False),
        NestedField(field_id=3, name="ask", field_type=DoubleType(), required=False),
    )

    # define a native Iceberg partition spec
    partition_spec = PartitionSpec(
        PartitionField(
            source_id=1,
            field_id=1000,
            transform=BucketTransform(2),
            name="symbol_bucket",
        )
    )

    # define a native Iceberg sort order
    # sort_order = SortOrder(SortField(source_id=1, transform=IdentityTransform()))

    # define the Daft dataframe to write
    df = daft.from_pydict(
        {
            "symbol": ["amzn", "goog", "meta", "msft"],
            "bid": [157.16, 150.55, 392.03, 403.25],
            "ask": [157.17, 150.56, 392.09, 403.27],
        }
    )

    # write to a table named `test_namespace.test_table_bucketed-<SUFFIX>`
    # we don't need to specify which catalog to create this table in since
    # only the "iceberg" catalog is available
    table_name = f"test_table_bucketed-{uuid.uuid4().hex[:8]}"
    namespace = "test_namespace"
    print(f"Creating Glue Table: {namespace}.{table_name}")
    dc.write_to_table(
        data=df,
        # path=warehouse + "/datafiles",
        table=table_name,
        namespace=namespace,
        schema=SchemaMapper.map(schema),
        partition_scheme=PartitionSchemeMapper.map(partition_spec, schema),
        # sort_keys=SortSchemeMapper.map(sort_order, schema),
    )

    print(f"Getting Glue Table: {namespace}.{table_name}")
    table_definition = dc.get_table(name=table_name, namespace=namespace)
    print(f"Retrieved Glue Table: {table_definition}")

    # Read Data from DeltaCAT Table (backed by Iceberg) using Daft
    daft_dataframe = dc.read_table(table=table_name, namespace=namespace)

    daft_dataframe.where(df["bid"] > 200.0).show()
    # Expected result:
    # ╭────────┬─────────┬─────────╮
    # │ symbol ┆ bid     ┆ ask     │
    # │ ---    ┆ ---     ┆ ---     │
    # │ Utf8   ┆ Float64 ┆ Float64 │
    # ╞════════╪═════════╪═════════╡
    # │ meta   ┆ 392.03  ┆ 392.09  │
    # ├╌╌╌╌╌╌╌╌┼╌╌╌╌╌╌╌╌╌┼╌╌╌╌╌╌╌╌╌┤
    # │ msft   ┆ 403.25  ┆ 403.27  │
    # ╰────────┴─────────┴─────────╯

    daft_dataframe.select("symbol").show()
    # Expected result:
    # ╭────────╮
    # │ symbol │
    # │ ---    │
    # │ Utf8   │
    # ╞════════╡
    # │ meta   │
    # ├╌╌╌╌╌╌╌╌┤
    # │ amzn   │
    # ├╌╌╌╌╌╌╌╌┤
    # │ goog   │
    # ├╌╌╌╌╌╌╌╌┤
    # │ msft   │
    # ╰────────╯


if __name__ == "__main__":
    example_script_args = [
        (
            [
                "--warehouse",
            ],
            {
                "help": "S3 path for Iceberg file storage.",
                "type": str,
            },
        ),
    ]

    # store any CLI args in the runtime environment
    store_cli_args_in_os_environ(example_script_args)

    # run the example using os.environ as kwargs
    run(**os.environ)



import pytest
from deltacat.storage import ManifestEntry, ManifestMeta
import pyarrow.parquet as papq
from deltacat.types.partial_download import PartialParquetParameters
from deltacat.compute.resource_estimation.manifest import (
    estimate_manifest_entry_column_size_bytes,
    estimate_manifest_entry_num_rows,
    estimate_manifest_entry_size_bytes,
)
from deltacat.compute.resource_estimation.model import (
    OperationType,
    EstimateResourcesParams,
    ResourceEstimationMethod,
)

PARQUET_FILE_PATH_NO_STATS = (
    "deltacat/tests/compute/resource_estimation/data/sample_no_stats.parquet"
)
PARQUET_FILE_PATH_WITH_STATS = (
    "deltacat/tests/compute/resource_estimation/data/sample_with_stats.parquet"
)


@pytest.fixture(scope="module")
def sample_no_stats_entry():
    manifest_meta = ManifestMeta.of(
        content_length=113629,
        record_count=0,
        content_type="application/parquet",
        content_encoding="identity",
        content_type_parameters=[
            PartialParquetParameters.of(
                pq_metadata=papq.ParquetFile(PARQUET_FILE_PATH_NO_STATS).metadata
            )
        ],
    )
    return ManifestEntry.of(
        url=PARQUET_FILE_PATH_NO_STATS,
        uri=PARQUET_FILE_PATH_NO_STATS,
        mandatory=True,
        uuid="test",
        meta=manifest_meta,
    )


@pytest.fixture(scope="module")
def sample_with_no_type_params():
    manifest_meta = ManifestMeta.of(
        content_length=113629,
        record_count=0,
        content_type="application/parquet",
        content_encoding="identity",
        content_type_parameters=[],
    )
    return ManifestEntry.of(
        url=PARQUET_FILE_PATH_NO_STATS,
        uri=PARQUET_FILE_PATH_NO_STATS,
        mandatory=True,
        uuid="test",
        meta=manifest_meta,
    )


@pytest.fixture(scope="module")
def sample_with_stats_entry():
    manifest_meta = ManifestMeta.of(
        content_length=113629,
        record_count=0,
        content_type="application/parquet",
        content_encoding="identity",
        content_type_parameters=[
            PartialParquetParameters.of(
                pq_metadata=papq.ParquetFile(PARQUET_FILE_PATH_WITH_STATS).metadata
            )
        ],
    )
    return ManifestEntry.of(
        url=PARQUET_FILE_PATH_WITH_STATS,
        uri=PARQUET_FILE_PATH_WITH_STATS,
        mandatory=True,
        uuid="test",
        meta=manifest_meta,
    )


class TestEstimateManifestEntryColumnSizeBytes:
    def test_when_no_columns_passed_sanity(
        self, sample_no_stats_entry, sample_with_stats_entry
    ):
        estimate_resources_params = EstimateResourcesParams.of(
            parquet_to_pyarrow_inflation=1
        )
        assert (
            int(
                estimate_manifest_entry_column_size_bytes(
                    sample_no_stats_entry,
                    operation_type=OperationType.PYARROW_DOWNLOAD,
                    estimate_resources_params=estimate_resources_params,
                )
            )
            == 0
        )
        assert (
            int(
                estimate_manifest_entry_column_size_bytes(
                    sample_with_stats_entry,
                    operation_type=OperationType.PYARROW_DOWNLOAD,
                    estimate_resources_params=estimate_resources_params,
                )
            )
            == 0
        )

    def test_when_no_columns_passed_with_intelligent_estimation(
        self, sample_no_stats_entry, sample_with_stats_entry
    ):
        estimate_resources_params = EstimateResourcesParams.of(
            parquet_to_pyarrow_inflation=1,
            resource_estimation_method=ResourceEstimationMethod.INTELLIGENT_ESTIMATION,
        )
        assert (
            int(
                estimate_manifest_entry_column_size_bytes(
                    sample_no_stats_entry,
                    operation_type=OperationType.PYARROW_DOWNLOAD,
                    estimate_resources_params=estimate_resources_params,
                )
            )
            == 0
        )
        assert (
            int(
                estimate_manifest_entry_column_size_bytes(
                    sample_with_stats_entry,
                    operation_type=OperationType.PYARROW_DOWNLOAD,
                    estimate_resources_params=estimate_resources_params,
                )
            )
            == 0
        )

    def test_when_one_string_column_passed(
        self, sample_no_stats_entry, sample_with_stats_entry
    ):
        estimate_resources_params = EstimateResourcesParams.of(
            parquet_to_pyarrow_inflation=1
        )
        assert (
            int(
                estimate_manifest_entry_column_size_bytes(
                    sample_no_stats_entry,
                    columns=["first_name"],
                    operation_type=OperationType.PYARROW_DOWNLOAD,
                    estimate_resources_params=estimate_resources_params,
                )
            )
            == 2988
        )
        assert (
            int(
                estimate_manifest_entry_column_size_bytes(
                    sample_with_stats_entry,
                    columns=["first_name"],
                    operation_type=OperationType.PYARROW_DOWNLOAD,
                    estimate_resources_params=estimate_resources_params,
                )
            )
            == 2989
        )

    def test_when_invalid_column_passed_assumes_null(self, sample_no_stats_entry):
        estimate_resources_params = EstimateResourcesParams.of(
            parquet_to_pyarrow_inflation=1
        )

        assert (
            int(
                estimate_manifest_entry_column_size_bytes(
                    sample_no_stats_entry,
                    columns=["invalid_column"],
                    operation_type=OperationType.PYARROW_DOWNLOAD,
                    estimate_resources_params=estimate_resources_params,
                )
            )
            == 4000
        )

    def test_when_multiple_columns_passed(
        self, sample_no_stats_entry, sample_with_stats_entry
    ):
        estimate_resources_params = EstimateResourcesParams.of(
            parquet_to_pyarrow_inflation=1
        )

        assert (
            int(
                estimate_manifest_entry_column_size_bytes(
                    sample_no_stats_entry,
                    columns=["first_name", "id"],
                    operation_type=OperationType.PYARROW_DOWNLOAD,
                    estimate_resources_params=estimate_resources_params,
                )
            )
            == 7031
        )
        assert (
            int(
                estimate_manifest_entry_column_size_bytes(
                    sample_with_stats_entry,
                    columns=["first_name", "id"],
                    operation_type=OperationType.PYARROW_DOWNLOAD,
                    estimate_resources_params=estimate_resources_params,
                )
            )
            == 8314
        )

    def test_when_timestamp_column_passed(
        self, sample_no_stats_entry, sample_with_stats_entry
    ):
        estimate_resources_params = EstimateResourcesParams.of(
            parquet_to_pyarrow_inflation=2
        )
        assert (
            int(
                estimate_manifest_entry_column_size_bytes(
                    sample_no_stats_entry,
                    columns=["registration_dttm"],
                    operation_type=OperationType.PYARROW_DOWNLOAD,
                    estimate_resources_params=estimate_resources_params,
                )
            )
            == 26540
        )
        assert (
            int(
                estimate_manifest_entry_column_size_bytes(
                    sample_with_stats_entry,
                    columns=["registration_dttm"],
                    operation_type=OperationType.PYARROW_DOWNLOAD,
                    estimate_resources_params=estimate_resources_params,
                )
            )
            == 18602
        )

    def test_when_intelligent_estimation_enabled_single_column(
        self, sample_no_stats_entry, sample_with_stats_entry
    ):
        estimate_resources_params = EstimateResourcesParams.of(
            parquet_to_pyarrow_inflation=1,
            resource_estimation_method=ResourceEstimationMethod.INTELLIGENT_ESTIMATION,
        )
        assert (
            int(
                estimate_manifest_entry_column_size_bytes(
                    sample_no_stats_entry,
                    columns=["first_name"],
                    operation_type=OperationType.PYARROW_DOWNLOAD,
                    estimate_resources_params=estimate_resources_params,
                )
            )
            == 2988
        )
        assert (
            int(
                estimate_manifest_entry_column_size_bytes(
                    sample_with_stats_entry,
                    columns=["first_name"],
                    operation_type=OperationType.PYARROW_DOWNLOAD,
                    estimate_resources_params=estimate_resources_params,
                )
            )
            == 7000
        )

    def test_when_intelligent_estimation_enabled_timestamp_column(
        self, sample_no_stats_entry, sample_with_stats_entry
    ):
        estimate_resources_params = EstimateResourcesParams.of(
            parquet_to_pyarrow_inflation=1,
            resource_estimation_method=ResourceEstimationMethod.INTELLIGENT_ESTIMATION,
        )
        assert (
            int(
                estimate_manifest_entry_column_size_bytes(
                    sample_no_stats_entry,
                    columns=["registration_dttm"],
                    operation_type=OperationType.PYARROW_DOWNLOAD,
                    estimate_resources_params=estimate_resources_params,
                )
            )
            == 12000
        )
        assert (
            int(
                estimate_manifest_entry_column_size_bytes(
                    sample_with_stats_entry,
                    columns=["registration_dttm"],
                    operation_type=OperationType.PYARROW_DOWNLOAD,
                    estimate_resources_params=estimate_resources_params,
                )
            )
            == 8000
        )

    def test_when_intelligent_estimation_enabled_int_column(
        self, sample_no_stats_entry, sample_with_stats_entry
    ):
        estimate_resources_params = EstimateResourcesParams.of(
            parquet_to_pyarrow_inflation=1,
            resource_estimation_method=ResourceEstimationMethod.INTELLIGENT_ESTIMATION,
        )
        assert (
            int(
                estimate_manifest_entry_column_size_bytes(
                    sample_no_stats_entry,
                    columns=["id"],
                    operation_type=OperationType.PYARROW_DOWNLOAD,
                    estimate_resources_params=estimate_resources_params,
                ),
            )
            == 4000
        )
        assert (
            int(
                estimate_manifest_entry_column_size_bytes(
                    sample_with_stats_entry,
                    columns=["id"],
                    operation_type=OperationType.PYARROW_DOWNLOAD,
                    estimate_resources_params=estimate_resources_params,
                )
            )
            == 4000
        )

    def test_when_intelligent_estimation_enabled_double_column(
        self, sample_no_stats_entry, sample_with_stats_entry
    ):
        estimate_resources_params = EstimateResourcesParams.of(
            parquet_to_pyarrow_inflation=1,
            resource_estimation_method=ResourceEstimationMethod.INTELLIGENT_ESTIMATION,
        )

        assert (
            int(
                estimate_manifest_entry_column_size_bytes(
                    sample_no_stats_entry,
                    columns=["salary"],
                    operation_type=OperationType.PYARROW_DOWNLOAD,
                    estimate_resources_params=estimate_resources_params,
                )
            )
            == 8000
        )
        assert (
            int(
                estimate_manifest_entry_column_size_bytes(
                    sample_with_stats_entry,
                    columns=["salary"],
                    operation_type=OperationType.PYARROW_DOWNLOAD,
                    estimate_resources_params=estimate_resources_params,
                )
            )
            == 8000
        )

    def test_when_intelligent_estimation_enabled_multiple_columns(
        self, sample_no_stats_entry, sample_with_stats_entry
    ):
        estimate_resources_params = EstimateResourcesParams.of(
            parquet_to_pyarrow_inflation=1,
            resource_estimation_method=ResourceEstimationMethod.INTELLIGENT_ESTIMATION,
        )

        assert (
            int(
                estimate_manifest_entry_column_size_bytes(
                    sample_no_stats_entry,
                    columns=["first_name", "id"],
                    operation_type=OperationType.PYARROW_DOWNLOAD,
                    estimate_resources_params=estimate_resources_params,
                )
            )
            == 6988
        )
        assert (
            int(
                estimate_manifest_entry_column_size_bytes(
                    sample_with_stats_entry,
                    columns=["first_name", "id"],
                    operation_type=OperationType.PYARROW_DOWNLOAD,
                    estimate_resources_params=estimate_resources_params,
                )
            )
            == 11000
        )

    def test_when_default_v2_enabled_multiple_columns(
        self, sample_no_stats_entry, sample_with_stats_entry
    ):
        estimate_resources_params = EstimateResourcesParams.of(
            parquet_to_pyarrow_inflation=1,
            resource_estimation_method=ResourceEstimationMethod.DEFAULT_V2,
        )

        assert (
            int(
                estimate_manifest_entry_column_size_bytes(
                    sample_no_stats_entry,
                    columns=["first_name", "id"],
                    operation_type=OperationType.PYARROW_DOWNLOAD,
                    estimate_resources_params=estimate_resources_params,
                )
            )
            == 6988
        )
        assert (
            int(
                estimate_manifest_entry_column_size_bytes(
                    sample_with_stats_entry,
                    columns=["first_name", "id"],
                    operation_type=OperationType.PYARROW_DOWNLOAD,
                    estimate_resources_params=estimate_resources_params,
                )
            )
            == 11000
        )

    def test_when_default_v2_enabled_multiple_columns_and_inflation_not_passed(
        self, sample_no_stats_entry
    ):
        estimate_resources_params = EstimateResourcesParams.of(
            parquet_to_pyarrow_inflation=None,
            resource_estimation_method=ResourceEstimationMethod.DEFAULT_V2,
        )

        assert (
            estimate_manifest_entry_column_size_bytes(
                sample_no_stats_entry,
                columns=["first_name", "id"],
                operation_type=OperationType.PYARROW_DOWNLOAD,
                estimate_resources_params=estimate_resources_params,
            )
            is None
        )

    def test_when_intelligent_estimation_enabled_with_no_type_params(
        self, sample_with_no_type_params
    ):
        estimate_resources_params = EstimateResourcesParams.of(
            parquet_to_pyarrow_inflation=1,
            resource_estimation_method=ResourceEstimationMethod.INTELLIGENT_ESTIMATION,
        )

        assert (
            estimate_manifest_entry_column_size_bytes(
                sample_with_no_type_params,
                columns=["first_name"],
                operation_type=OperationType.PYARROW_DOWNLOAD,
                estimate_resources_params=estimate_resources_params,
            )
            is None
        )

    def test_when_previous_inflation_method_with_no_type_params(
        self, sample_with_no_type_params
    ):
        estimate_resources_params = EstimateResourcesParams.of(
            parquet_to_pyarrow_inflation=1,
            resource_estimation_method=ResourceEstimationMethod.PREVIOUS_INFLATION,
        )

        assert (
            estimate_manifest_entry_column_size_bytes(
                sample_with_no_type_params,
                columns=["first_name"],
                operation_type=OperationType.PYARROW_DOWNLOAD,
                estimate_resources_params=estimate_resources_params,
            )
            is None
        )

    def test_when_default_v2_method_with_no_type_params(
        self, sample_with_no_type_params
    ):
        estimate_resources_params = EstimateResourcesParams.of(
            parquet_to_pyarrow_inflation=1,
            resource_estimation_method=ResourceEstimationMethod.DEFAULT_V2,
        )

        assert (
            estimate_manifest_entry_column_size_bytes(
                sample_with_no_type_params,
                columns=["first_name"],
                operation_type=OperationType.PYARROW_DOWNLOAD,
                estimate_resources_params=estimate_resources_params,
            )
            is None
        )


class TestEstimateManifestEntryNumRows:
    def test_sanity(self, sample_no_stats_entry):
        estimate_resources_params = EstimateResourcesParams.of(
            parquet_to_pyarrow_inflation=1,
            previous_inflation=7,
            average_record_size_bytes=1000,
            resource_estimation_method=ResourceEstimationMethod.DEFAULT,
        )

        assert (
            estimate_manifest_entry_num_rows(
                sample_no_stats_entry,
                operation_type=OperationType.PYARROW_DOWNLOAD,
                estimate_resources_params=estimate_resources_params,
            )
            == 1000
        )

    def test_when_previous_inflation_forced(self, sample_no_stats_entry):
        estimate_resources_params = EstimateResourcesParams.of(
            parquet_to_pyarrow_inflation=1,
            previous_inflation=7,
            average_record_size_bytes=1000,
            resource_estimation_method=ResourceEstimationMethod.PREVIOUS_INFLATION,
        )
        assert (
            estimate_manifest_entry_num_rows(
                sample_no_stats_entry,
                operation_type=OperationType.PYARROW_DOWNLOAD,
                estimate_resources_params=estimate_resources_params,
            )
            == 795
        )

    def test_when_type_params_absent_default_method(self, sample_with_no_type_params):
        estimate_resources_params = EstimateResourcesParams.of(
            parquet_to_pyarrow_inflation=1,
            previous_inflation=7,
            average_record_size_bytes=1000,
            resource_estimation_method=ResourceEstimationMethod.DEFAULT,
        )

        assert (
            estimate_manifest_entry_num_rows(
                sample_with_no_type_params,
                operation_type=OperationType.PYARROW_DOWNLOAD,
                estimate_resources_params=estimate_resources_params,
            )
            == 795
        )

    def test_when_type_params_absent_intelligent_estimation(
        self, sample_with_no_type_params
    ):
        estimate_resources_params = EstimateResourcesParams.of(
            parquet_to_pyarrow_inflation=1,
            previous_inflation=7,
            average_record_size_bytes=1000,
            resource_estimation_method=ResourceEstimationMethod.INTELLIGENT_ESTIMATION,
        )

        assert (
            estimate_manifest_entry_num_rows(
                sample_with_no_type_params,
                operation_type=OperationType.PYARROW_DOWNLOAD,
                estimate_resources_params=estimate_resources_params,
            )
            is None
        )

    def test_when_type_params_absent_content_type_meta(
        self, sample_with_no_type_params
    ):
        estimate_resources_params = EstimateResourcesParams.of(
            parquet_to_pyarrow_inflation=1,
            previous_inflation=7,
            average_record_size_bytes=1000,
            resource_estimation_method=ResourceEstimationMethod.CONTENT_TYPE_META,
        )

        assert (
            estimate_manifest_entry_num_rows(
                sample_with_no_type_params,
                operation_type=OperationType.PYARROW_DOWNLOAD,
                estimate_resources_params=estimate_resources_params,
            )
            is None
        )

    def test_when_type_params_absent_previous_inflation(
        self, sample_with_no_type_params
    ):
        estimate_resources_params = EstimateResourcesParams.of(
            parquet_to_pyarrow_inflation=1,
            previous_inflation=7,
            average_record_size_bytes=1000,
            resource_estimation_method=ResourceEstimationMethod.PREVIOUS_INFLATION,
        )

        assert (
            estimate_manifest_entry_num_rows(
                sample_with_no_type_params,
                operation_type=OperationType.PYARROW_DOWNLOAD,
                estimate_resources_params=estimate_resources_params,
            )
            == 795
        )

    def test_when_type_params_absent_default_v2(self, sample_with_no_type_params):
        estimate_resources_params = EstimateResourcesParams.of(
            parquet_to_pyarrow_inflation=1,
            previous_inflation=7,
            average_record_size_bytes=1000,
            resource_estimation_method=ResourceEstimationMethod.DEFAULT_V2,
        )

        assert (
            estimate_manifest_entry_num_rows(
                sample_with_no_type_params,
                operation_type=OperationType.PYARROW_DOWNLOAD,
                estimate_resources_params=estimate_resources_params,
            )
            == 795  # same as previous inflation
        )

    def test_when_type_params_no_stats_with_default_v2(self, sample_no_stats_entry):
        estimate_resources_params = EstimateResourcesParams.of(
            parquet_to_pyarrow_inflation=1,
            previous_inflation=7,
            average_record_size_bytes=1000,
            resource_estimation_method=ResourceEstimationMethod.DEFAULT_V2,
        )
        assert (
            estimate_manifest_entry_num_rows(
                sample_no_stats_entry,
                operation_type=OperationType.PYARROW_DOWNLOAD,
                estimate_resources_params=estimate_resources_params,
            )
            == 1000
        )

    def test_when_type_params_parquet_inflation_absent_with_default_v2(
        self, sample_no_stats_entry
    ):
        estimate_resources_params = EstimateResourcesParams.of(
            parquet_to_pyarrow_inflation=None,
            previous_inflation=7,
            average_record_size_bytes=1000,
            resource_estimation_method=ResourceEstimationMethod.DEFAULT_V2,
        )
        assert (
            estimate_manifest_entry_num_rows(
                sample_no_stats_entry,
                operation_type=OperationType.PYARROW_DOWNLOAD,
                estimate_resources_params=estimate_resources_params,
            )
            == 1000
        )

    def test_when_type_params_with_default_v2(self, sample_with_stats_entry):
        estimate_resources_params = EstimateResourcesParams.of(
            parquet_to_pyarrow_inflation=1,
            previous_inflation=7,
            average_record_size_bytes=1000,
            resource_estimation_method=ResourceEstimationMethod.DEFAULT_V2,
        )
        assert (
            estimate_manifest_entry_num_rows(
                sample_with_stats_entry,
                operation_type=OperationType.PYARROW_DOWNLOAD,
                estimate_resources_params=estimate_resources_params,
            )
            == 1000
        )


class TestEstimateManifestEntrySizeBytes:
    def test_sanity(self, sample_no_stats_entry):
        estimate_resources_params = EstimateResourcesParams.of(
            parquet_to_pyarrow_inflation=2,
            previous_inflation=7,
            resource_estimation_method=ResourceEstimationMethod.DEFAULT,
        )

        assert (
            int(
                estimate_manifest_entry_size_bytes(
                    sample_no_stats_entry,
                    operation_type=OperationType.PYARROW_DOWNLOAD,
                    estimate_resources_params=estimate_resources_params,
                )
            )
            == 224984
        )

    def test_when_previous_inflation_forced(self, sample_no_stats_entry):
        estimate_resources_params = EstimateResourcesParams.of(
            parquet_to_pyarrow_inflation=2,
            previous_inflation=7,
            resource_estimation_method=ResourceEstimationMethod.PREVIOUS_INFLATION,
        )
        assert (
            int(
                estimate_manifest_entry_size_bytes(
                    sample_no_stats_entry,
                    operation_type=OperationType.PYARROW_DOWNLOAD,
                    estimate_resources_params=estimate_resources_params,
                )
            )
            == 795403
        )

    def test_when_type_params_absent_default(self, sample_with_no_type_params):
        estimate_resources_params = EstimateResourcesParams.of(
            parquet_to_pyarrow_inflation=2,
            previous_inflation=7,
            resource_estimation_method=ResourceEstimationMethod.DEFAULT,
        )
        assert (
            int(
                estimate_manifest_entry_size_bytes(
                    sample_with_no_type_params,
                    operation_type=OperationType.PYARROW_DOWNLOAD,
                    estimate_resources_params=estimate_resources_params,
                )
            )
            == 795403
        )

    def test_when_type_params_absent_intelligent_estimation(
        self, sample_with_no_type_params
    ):
        estimate_resources_params = EstimateResourcesParams.of(
            parquet_to_pyarrow_inflation=2,
            previous_inflation=7,
            resource_estimation_method=ResourceEstimationMethod.INTELLIGENT_ESTIMATION,
        )
        assert (
            estimate_manifest_entry_size_bytes(
                sample_with_no_type_params,
                operation_type=OperationType.PYARROW_DOWNLOAD,
                estimate_resources_params=estimate_resources_params,
            )
            is None
        )

    def test_when_type_params_absent_content_meta(self, sample_with_no_type_params):
        estimate_resources_params = EstimateResourcesParams.of(
            parquet_to_pyarrow_inflation=2,
            previous_inflation=7,
            resource_estimation_method=ResourceEstimationMethod.CONTENT_TYPE_META,
        )
        assert (
            estimate_manifest_entry_size_bytes(
                sample_with_no_type_params,
                operation_type=OperationType.PYARROW_DOWNLOAD,
                estimate_resources_params=estimate_resources_params,
            )
            is None
        )

    def test_when_type_params_absent_previous_inflation(
        self, sample_with_no_type_params
    ):
        estimate_resources_params = EstimateResourcesParams.of(
            parquet_to_pyarrow_inflation=2,
            previous_inflation=7,
            resource_estimation_method=ResourceEstimationMethod.PREVIOUS_INFLATION,
        )
        assert (
            estimate_manifest_entry_size_bytes(
                sample_with_no_type_params,
                operation_type=OperationType.PYARROW_DOWNLOAD,
                estimate_resources_params=estimate_resources_params,
            )
            == 795403
        )

    def test_when_intelligent_estimation_sanity(self, sample_no_stats_entry):
        estimate_resources_params = EstimateResourcesParams.of(
            parquet_to_pyarrow_inflation=2,
            previous_inflation=7,
            resource_estimation_method=ResourceEstimationMethod.INTELLIGENT_ESTIMATION,
        )
        assert (
            int(
                estimate_manifest_entry_size_bytes(
                    sample_no_stats_entry,
                    operation_type=OperationType.PYARROW_DOWNLOAD,
                    estimate_resources_params=estimate_resources_params,
                )
            )
            == 223096
        )

    def test_when_type_params_with_stats_default_method(self, sample_with_stats_entry):
        estimate_resources_params = EstimateResourcesParams.of(
            parquet_to_pyarrow_inflation=2,
            previous_inflation=7,
            resource_estimation_method=ResourceEstimationMethod.DEFAULT,
        )
        assert (
            int(
                estimate_manifest_entry_size_bytes(
                    sample_with_stats_entry,
                    operation_type=OperationType.PYARROW_DOWNLOAD,
                    estimate_resources_params=estimate_resources_params,
                )
            )
            == 227794
        )

    def test_when_type_params_with_stats_intelligent_method(
        self, sample_with_stats_entry
    ):
        estimate_resources_params = EstimateResourcesParams.of(
            parquet_to_pyarrow_inflation=2,
            previous_inflation=7,
            resource_estimation_method=ResourceEstimationMethod.INTELLIGENT_ESTIMATION,
        )
        assert (
            int(
                estimate_manifest_entry_size_bytes(
                    sample_with_stats_entry,
                    operation_type=OperationType.PYARROW_DOWNLOAD,
                    estimate_resources_params=estimate_resources_params,
                )
            )
            == 290222
        )

    def test_when_type_params_with_content_type_meta_method(
        self, sample_with_stats_entry
    ):
        estimate_resources_params = EstimateResourcesParams.of(
            parquet_to_pyarrow_inflation=2,
            previous_inflation=7,
            resource_estimation_method=ResourceEstimationMethod.CONTENT_TYPE_META,
        )
        assert (
            int(
                estimate_manifest_entry_size_bytes(
                    sample_with_stats_entry,
                    operation_type=OperationType.PYARROW_DOWNLOAD,
                    estimate_resources_params=estimate_resources_params,
                )
            )
            == 227794
        )

    def test_when_type_params_with_stats_default_v2_method(
        self, sample_with_stats_entry
    ):
        estimate_resources_params = EstimateResourcesParams.of(
            parquet_to_pyarrow_inflation=2,
            previous_inflation=7,
            resource_estimation_method=ResourceEstimationMethod.DEFAULT_V2,
        )
        assert (
            int(
                estimate_manifest_entry_size_bytes(
                    sample_with_stats_entry,
                    operation_type=OperationType.PYARROW_DOWNLOAD,
                    estimate_resources_params=estimate_resources_params,
                )
            )
            == 290222  # same result as intelligent estimation
        )

    def test_when_type_params_without_stats_default_v2_method(
        self, sample_no_stats_entry
    ):
        estimate_resources_params = EstimateResourcesParams.of(
            parquet_to_pyarrow_inflation=2,
            previous_inflation=7,
            resource_estimation_method=ResourceEstimationMethod.DEFAULT_V2,
        )
        assert (
            int(
                estimate_manifest_entry_size_bytes(
                    sample_no_stats_entry,
                    operation_type=OperationType.PYARROW_DOWNLOAD,
                    estimate_resources_params=estimate_resources_params,
                )
            )
            == 223096
        )

    def test_when_no_type_params_default_v2_method(self, sample_with_no_type_params):
        estimate_resources_params = EstimateResourcesParams.of(
            parquet_to_pyarrow_inflation=2,
            previous_inflation=7,
            resource_estimation_method=ResourceEstimationMethod.DEFAULT_V2,
        )
        assert (
            estimate_manifest_entry_size_bytes(
                sample_with_no_type_params,
                operation_type=OperationType.PYARROW_DOWNLOAD,
                estimate_resources_params=estimate_resources_params,
            )
            == 795403  # same as previous inflation
        )

    def test_when_type_params_but_inflation_absent_default_v2_method(
        self, sample_with_stats_entry
    ):
        estimate_resources_params = EstimateResourcesParams.of(
            parquet_to_pyarrow_inflation=None,
            previous_inflation=7,
            resource_estimation_method=ResourceEstimationMethod.DEFAULT_V2,
        )
        assert (
            estimate_manifest_entry_size_bytes(
                sample_with_stats_entry,
                operation_type=OperationType.PYARROW_DOWNLOAD,
                estimate_resources_params=estimate_resources_params,
            )
            == 795403  # same as previous inflation
        )



import math
import os
from enum import Enum
from io import BytesIO
from typing import Any, Dict

import faker
from faker_file.providers.png_file import GraphicPngFileProvider
from faker_file.storages.filesystem import FileSystemStorage
from PIL import Image

from deltacat.benchmarking.data.row_generator import RowGenerator


class ImageStyle(Enum):
    RANDOM_BYTES = 1
    """Generate random bytes to simulate an image.

    This is the fastest option (if you want to test correctness).
    """
    PILLOW = 2
    """Generate actual PNG files in-memory directly using Pillow"""
    FAKER_FILE = 3
    """Generate PNG files on-disk using with some random elements"""


class RandomRowGenerator(RowGenerator):
    """Generate rows with 'images' that are just randomly-generated bytes"""

    def __init__(
        self, seed=0, tmp_dir=None, style: ImageStyle = ImageStyle.RANDOM_BYTES
    ):
        self.seed = seed
        self.fake = faker.Faker()
        self.fake.seed_instance(seed)
        self.fake.add_provider(GraphicPngFileProvider)
        self.temp_dir = tmp_dir
        self.style = style

    def __str__(self):
        return f"random source"

    def _generate_image(self, width, height) -> bytes:
        if self.style == ImageStyle.RANDOM_BYTES:
            return self._generate_with_random_bytes(width, height)
        elif self.style == ImageStyle.PILLOW:
            return self._generate_with_pillow(width, height)
        elif self.style == ImageStyle.FAKER_FILE:
            return self._generate_with_faker(width, height)
        else:
            raise ValueError("Unknown ImageStyle")

    @staticmethod
    def _generate_with_random_bytes(width, height) -> bytes:
        """Generate random bytes to simulate an image."""
        target_size = math.floor(
            width * height / 50
        )  # this isn't actually how file size relates to image size
        # Assumption: we don't actually need images. It suffices to generate arbitrary-length bytes of random characters.
        return os.urandom(target_size)

    @staticmethod
    def _generate_with_pillow(width, height) -> bytes:
        """Generate actual PNG files in-memory directly using Pillow"""
        file = BytesIO()
        image = Image.new("RGBA", size=(width, height), color=(155, 0, 0))
        image.save(file, "png")
        file.name = "test.png"
        file.seek(0)
        return file.read()

    def _generate_with_faker(self, width, height) -> bytes:
        """Generate PNG files on-disk using with some random elements"""
        rel_name = self.fake.graphic_png_file(
            storage=FileSystemStorage(
                root_path=self.temp_dir,
                rel_path="tmp",
            ),
            size=(width, height),
        )
        file_name = f"{self.temp_dir}/{rel_name}"
        with open(file_name, "rb") as f:
            return f.read()

    def generate(self) -> Dict[str, Any]:
        return {
            "id": self.fake.random_int(0, 10_000_000),
            "source": self.fake.image_url(),
            "media": (
                self._generate_image(
                    self.fake.random_int(512, 2048), self.fake.random_int(512, 4096)
                )
            ),
        }



from deltacat.storage.model.manifest import (
    EntryType,
    EntryParams,
    Manifest,
    ManifestAuthor,
    ManifestEntry,
    ManifestEntryList,
    ManifestMeta,
)
from deltacat.storage.model.delta import (
    Delta,
    DeltaLocator,
    DeltaProperties,
)
from deltacat.storage.model.list_result import ListResult
from deltacat.storage.model.locator import Locator
from deltacat.storage.model.metafile import (
    Metafile,
)
from deltacat.storage.model.transaction import (
    TransactionOperation,
    Transaction,
    read_transaction,
    transactions,
    transaction,
)
from deltacat.storage.model.namespace import (
    Namespace,
    NamespaceLocator,
    NamespaceProperties,
)
from deltacat.storage.model.partition import (
    Partition,
    PartitionLocator,
    PartitionLocatorAlias,
    PartitionKey,
    PartitionKeyList,
    PartitionScheme,
    PartitionSchemeList,
    PartitionValues,
)
from deltacat.storage.model.schema import (
    Field,
    FieldId,
    FieldLocator,
    FieldName,
    NestedFieldName,
    Schema,
    SchemaList,
    SchemaUpdate,
    SchemaUpdateOperation,
    SchemaUpdateOperations,
)
from deltacat.storage.model.stream import (
    Stream,
    StreamLocator,
    StreamLocatorAlias,
)
from deltacat.storage.model.table import (
    Table,
    TableLocator,
    TableProperties,
)
from deltacat.storage.model.table_version import (
    TableVersion,
    TableVersionLocator,
    TableVersionProperties,
)
from deltacat.storage.model.transform import (
    Transform,
    TransformName,
    TransformParameters,
    BucketingStrategy,
    BucketTransformParameters,
    TruncateTransformParameters,
    BucketTransform,
    IdentityTransform,
    VoidTransform,
    UnknownTransform,
    HourTransform,
    DayTransform,
    MonthTransform,
    YearTransform,
    TruncateTransform,
    TruncateStrategy,
)
from deltacat.storage.model.types import (
    CommitState,
    Dataset,
    DeltaType,
    DistributedDataset,
    LifecycleState,
    LocalDataset,
    LocalTable,
    NullOrder,
    SchemaConsistencyType,
    StreamFormat,
    SortOrder,
    TransactionOperationType,
    TransactionStatus,
)
from deltacat.storage.model.sort_key import (
    SortKey,
    SortKeyList,
    SortScheme,
    SortSchemeList,
)
from deltacat.storage.main import impl as metastore

__all__ = [
    "BucketingStrategy",
    "BucketTransform",
    "BucketTransformParameters",
    "CommitState",
    "Dataset",
    "DayTransform",
    "Delta",
    "DeltaLocator",
    "DeltaProperties",
    "DeltaType",
    "DistributedDataset",
    "EntryType",
    "EntryParams",
    "Field",
    "FieldId",
    "FieldLocator",
    "FieldName",
    "HourTransform",
    "IdentityTransform",
    "LifecycleState",
    "ListResult",
    "LocalDataset",
    "LocalTable",
    "Locator",
    "Manifest",
    "ManifestAuthor",
    "ManifestEntry",
    "ManifestEntryList",
    "ManifestMeta",
    "Metafile",
    "metastore",
    "MonthTransform",
    "Namespace",
    "NamespaceLocator",
    "NamespaceProperties",
    "NestedFieldName",
    "NullOrder",
    "Partition",
    "PartitionKey",
    "PartitionKeyList",
    "PartitionLocator",
    "PartitionLocatorAlias",
    "PartitionScheme",
    "PartitionSchemeList",
    "PartitionValues",
    "Schema",
    "SchemaList",
    "SchemaUpdate",
    "SchemaUpdateOperation",
    "SchemaUpdateOperations",
    "SchemaConsistencyType",
    "SortKey",
    "SortKeyList",
    "SortOrder",
    "SortScheme",
    "SortSchemeList",
    "Stream",
    "StreamFormat",
    "StreamLocator",
    "StreamLocatorAlias",
    "Table",
    "TableLocator",
    "TableProperties",
    "TableVersion",
    "TableVersionLocator",
    "TableVersionProperties",
    "Transaction",
    "TransactionOperation",
    "TransactionOperationType",
    "TransactionStatus",
    "Transform",
    "TransformName",
    "TransformParameters",
    "TruncateTransform",
    "TruncateTransformParameters",
    "TruncateStrategy",
    "UnknownTransform",
    "VoidTransform",
    "YearTransform",
    "read_transaction",
    "transactions",
    "transaction",
]



from deltacat.experimental.catalog.iceberg.iceberg_catalog_config import (
    IcebergCatalogConfig,
)
import deltacat.experimental.catalog.iceberg.impl as IcebergCatalog

__all__ = ["IcebergCatalogConfig", "IcebergCatalog"]



import logging
import os
import json
import pathlib
from logging import FileHandler, Handler, Logger, LoggerAdapter, handlers
from typing import Any, Dict, Optional, Union

import ray
from ray.runtime_context import RuntimeContext

from deltacat.constants import (
    DELTACAT_APP_LOG_LEVEL,
    DELTACAT_SYS_LOG_LEVEL,
    DELTACAT_APP_LOG_DIR,
    DELTACAT_SYS_LOG_DIR,
    DELTACAT_APP_INFO_LOG_BASE_FILE_NAME,
    DELTACAT_SYS_INFO_LOG_BASE_FILE_NAME,
    DELTACAT_APP_DEBUG_LOG_BASE_FILE_NAME,
    DELTACAT_SYS_DEBUG_LOG_BASE_FILE_NAME,
    DELTACAT_LOGGER_CONTEXT,
    DELTACAT_LOGGER_USE_SINGLE_HANDLER,
)

DEFAULT_LOG_LEVEL = "INFO"
DEFAULT_LOG_FORMAT = {
    "level": "levelname",
    "message": "message",
    "loggerName": "name",
    "processName": "processName",
    "processID": "process",
    "threadName": "threadName",
    "timestamp": "asctime",
    "filename": "filename",
    "lineno": "lineno",
}
DEFAULT_MAX_BYTES_PER_LOG = 2 ^ 20 * 256  # 256 MiB
DEFAULT_BACKUP_COUNT = 0


class JsonFormatter(logging.Formatter):
    """
    Formatter that outputs JSON strings after parsing the LogRecord.

    @param dict fmt_dict: Key: logging format attribute pairs. Defaults to {"message": "message"}.
    @param str time_format: time.strftime() format string. Default: "%Y-%m-%dT%H:%M:%S"
    @param str msec_format: Microsecond formatting. Appended at the end. Default: "%s.%03dZ"
    """

    def __init__(
        self,
        fmt_dict: dict = None,
        time_format: str = "%Y-%m-%dT%H:%M:%S",
        msec_format: str = "%s.%03dZ",
        context_kwargs: Optional[Dict[str, Any]] = None,
    ):
        self.fmt_dict = fmt_dict if fmt_dict is not None else {"message": "message"}
        self.default_time_format = time_format
        self.default_msec_format = msec_format
        self.datefmt = None
        self.additional_context = context_kwargs or {}
        if ray.is_initialized():
            self.ray_runtime_ctx: RuntimeContext = ray.get_runtime_context()
            self.context = {}
            self.context["worker_id"] = self.ray_runtime_ctx.get_worker_id()
            self.context["node_id"] = self.ray_runtime_ctx.get_node_id()
            self.context["job_id"] = self.ray_runtime_ctx.get_job_id()
        else:
            self.ray_runtime_ctx = None
            self.context = {}

        if DELTACAT_LOGGER_CONTEXT is not None:
            try:
                env_context = json.loads(DELTACAT_LOGGER_CONTEXT)
                self.additional_context.update(env_context)
            except Exception:
                pass

    def usesTime(self) -> bool:
        """
        Overwritten to look for the attribute in the format dict values instead of the fmt string.
        """
        return "asctime" in self.fmt_dict.values()

    def formatMessage(self, record) -> dict:
        """
        Overwritten to return a dictionary of the relevant LogRecord attributes instead of a string.
        KeyError is raised if an unknown attribute is provided in the fmt_dict.
        """
        return {
            fmt_key: record.__dict__[fmt_val]
            for fmt_key, fmt_val in self.fmt_dict.items()
        }

    def format(self, record) -> str:
        """
        Mostly the same as the parent's class method, the difference being that a dict is manipulated and dumped as JSON
        instead of a string.
        """
        record.message = record.getMessage()

        if self.usesTime():
            record.asctime = self.formatTime(record, self.datefmt)

        message_dict = self.formatMessage(record)

        if record.exc_info:
            # Cache the traceback text to avoid converting it multiple times
            # (it's constant anyway)
            if not record.exc_text:
                record.exc_text = self.formatException(record.exc_info)

        if record.exc_text:
            message_dict["exc_info"] = record.exc_text

        if record.stack_info:
            message_dict["stack_info"] = self.formatStack(record.stack_info)

        if self.ray_runtime_ctx:
            # only workers will have task ID
            if (
                self.ray_runtime_ctx.worker
                and self.ray_runtime_ctx.worker.mode == ray._private.worker.WORKER_MODE
            ):
                self.context["task_id"] = self.ray_runtime_ctx.get_task_id()
                self.context[
                    "assigned_resources"
                ] = self.ray_runtime_ctx.get_assigned_resources()

            message_dict["ray_runtime_context"] = self.context

        if self.additional_context:
            message_dict["additional_context"] = self.additional_context

        return json.dumps(message_dict, default=str)


class DeltaCATLoggerAdapter(logging.LoggerAdapter):
    """
    Logger Adapter class with additional functionality
    """

    def __init__(self, logger: Logger, extra: Optional[Dict[str, Any]] = {}):
        super().__init__(logger, extra)

    def debug_conditional(self, msg, do_print: bool, *args, **kwargs):
        if do_print:
            self.debug(msg, *args, **kwargs)

    def info_conditional(self, msg, do_print: bool, *args, **kwargs):
        if do_print:
            self.info(msg, *args, **kwargs)

    def warning_conditional(self, msg, do_print: bool, *args, **kwargs):
        if do_print:
            self.warning(msg, *args, **kwargs)

    def error_conditional(self, msg, do_print: bool, *args, **kwargs):
        if do_print:
            self.error(msg, *args, **kwargs)


def _add_logger_handler(logger: Logger, handler: Handler) -> Logger:

    logger.setLevel(logging.getLevelName("DEBUG"))
    logger.addHandler(handler)
    return logger


def _create_rotating_file_handler(
    log_directory: str,
    log_base_file_name: str,
    logging_level: Union[str, int] = DEFAULT_LOG_LEVEL,
    max_bytes_per_log_file: int = DEFAULT_MAX_BYTES_PER_LOG,
    backup_count: int = DEFAULT_BACKUP_COUNT,
    logging_format: Union[str, dict] = DEFAULT_LOG_FORMAT,
    context_kwargs: Dict[str, Any] = None,
) -> FileHandler:

    if type(logging_level) is str:
        logging_level = logging.getLevelName(logging_level.upper())
    assert log_base_file_name, "log file name is required"
    assert log_directory, "log directory is required"
    log_dir_path = pathlib.Path(log_directory)
    log_dir_path.mkdir(parents=True, exist_ok=True)
    handler = handlers.RotatingFileHandler(
        os.path.join(log_directory, log_base_file_name),
        maxBytes=max_bytes_per_log_file,
        backupCount=backup_count,
    )

    if type(logging_format) is str:
        handler.setFormatter(logging.Formatter(logging_format))
    else:
        handler.setFormatter(
            JsonFormatter(logging_format, context_kwargs=context_kwargs)
        )

    handler.setLevel(logging_level)
    return handler


def _file_handler_exists(logger: Logger, log_dir: str, log_base_file_name: str) -> bool:

    handler_exists = False
    base_file_path = os.path.join(log_dir, log_base_file_name)

    if logger.handlers:
        norm_base_file_path = os.path.normpath(base_file_path)
        handler_exists = any(
            [
                isinstance(handler, logging.FileHandler)
                and os.path.normpath(handler.baseFilename) == norm_base_file_path
                for handler in logger.handlers
            ]
        )
    return handler_exists


def _configure_logger(
    logger: Logger,
    log_level: int,
    log_dir: str,
    log_base_file_name: str,
    debug_log_base_file_name: str,
    context_kwargs: Dict[str, Any] = None,
) -> Union[Logger, LoggerAdapter]:
    # This maintains log level of rotating file handlers
    primary_log_level = log_level
    logger.propagate = False
    needs_handler = True
    if log_level <= logging.getLevelName("DEBUG"):
        if not _file_handler_exists(logger, log_dir, debug_log_base_file_name):
            handler = _create_rotating_file_handler(
                log_dir,
                debug_log_base_file_name,
                "DEBUG",
                context_kwargs=context_kwargs,
            )
            _add_logger_handler(logger, handler)
            needs_handler = not DELTACAT_LOGGER_USE_SINGLE_HANDLER
            primary_log_level = logging.getLevelName("INFO")
    if not _file_handler_exists(logger, log_dir, log_base_file_name) and needs_handler:
        handler = _create_rotating_file_handler(
            log_dir,
            log_base_file_name,
            primary_log_level,
            context_kwargs=context_kwargs,
        )
        _add_logger_handler(logger, handler)

    return DeltaCATLoggerAdapter(logger)


def configure_deltacat_logger(
    logger: Logger,
    level: int = None,
    context_kwargs: Dict[str, Any] = None,
) -> Union[Logger, LoggerAdapter]:
    if level is None:
        level = logging.getLevelName(DELTACAT_SYS_LOG_LEVEL)

    return _configure_logger(
        logger,
        level,
        DELTACAT_SYS_LOG_DIR,
        DELTACAT_SYS_INFO_LOG_BASE_FILE_NAME,
        DELTACAT_SYS_DEBUG_LOG_BASE_FILE_NAME,
        context_kwargs,
    )


def configure_application_logger(
    logger: Logger,
    level: int = None,
    context_kwargs: Dict[str, Any] = None,
) -> Union[Logger, LoggerAdapter]:
    if level is None:
        level = logging.getLevelName(DELTACAT_APP_LOG_LEVEL)

    return _configure_logger(
        logger,
        level,
        DELTACAT_APP_LOG_DIR,
        DELTACAT_APP_INFO_LOG_BASE_FILE_NAME,
        DELTACAT_APP_DEBUG_LOG_BASE_FILE_NAME,
        context_kwargs,
    )



#!/usr/bin/env python3
"""
DeltaCAT Compactor Bootstrap Script

This script creates test data suitable for compaction testing by:
1. Creating source and destination namespaces and tables with schema
2. Writing 2 test parquet files as separate deltas to the source table
3. Staging and committing all necessary deltacat metadata (table version, stream, partition, deltas)
4. Running compaction using the direct API (not the CLI script)

Usage:
    # Use default catalog location
    python bootstrap.py

    # Use custom catalog location
    python bootstrap.py --catalog-root /path/to/catalog

    # Automatically run compaction after bootstrapping
    python bootstrap.py --run-compaction yes

    # Automatically run compaction after bootstrapping
    python bootstrap.py --run-compaction yes

    # Automatically run compaction against an S3 catalog (bucket must exist)
    python bootstrap.py --run-compaction yes --catalog-root s3://bucket/key

The script creates:
- A source namespace "compactor_test_source"
- A destination namespace "compactor_test_dest"
- Source table "events" with columns: id, timestamp, user_id, event_type, data
- Destination table "events_compacted"
- 2 parquet files with overlapping data (suitable for compaction)
- All necessary deltacat metadata (table version, stream, partition, deltas)
- Working end-to-end compaction demonstration
"""

import argparse
import sys

import pandas as pd

from deltacat.catalog import write_to_table, get_table, create_table
from deltacat.types.media import ContentType
from deltacat.storage import metastore
from deltacat.types.tables import TableWriteMode

# Import common utilities
from deltacat.examples.compactor.utils.common import (
    get_default_catalog_root,
    initialize_catalog,
)


def create_test_data_batch_1() -> pd.DataFrame:
    """Create the first batch of test data with some overlapping IDs."""
    return pd.DataFrame(
        {
            "id": [1, 2, 3, 4, 5],
            "timestamp": pd.to_datetime(
                [
                    "2024-01-01 10:00:00",
                    "2024-01-01 10:05:00",
                    "2024-01-01 10:10:00",
                    "2024-01-01 10:15:00",
                    "2024-01-01 10:20:00",
                ]
            ),
            "user_id": [101, 102, 103, 104, 105],
            "event_type": ["login", "view", "click", "purchase", "logout"],
            "data": [
                '{"page": "home"}',
                '{"product_id": 123}',
                '{"button": "add_to_cart"}',
                '{"amount": 99.99}',
                '{"session_duration": 1200}',
            ],
        }
    )


def create_test_data_batch_2() -> pd.DataFrame:
    """Create the second batch of test data with some overlapping IDs (good for compaction)."""
    return pd.DataFrame(
        {
            "id": [3, 4, 5, 6, 7, 8],  # IDs 3, 4, 5 overlap with batch 1
            "timestamp": pd.to_datetime(
                [
                    "2024-01-01 11:00:00",  # Later timestamp for ID 3 (should replace)
                    "2024-01-01 11:05:00",  # Later timestamp for ID 4 (should replace)
                    "2024-01-01 11:10:00",  # Later timestamp for ID 5 (should replace)
                    "2024-01-01 11:15:00",  # New ID 6
                    "2024-01-01 11:20:00",  # New ID 7
                    "2024-01-01 11:25:00",  # New ID 8
                ]
            ),
            "user_id": [103, 104, 105, 106, 107, 108],
            "event_type": ["view", "click", "purchase", "login", "view", "logout"],
            "data": [
                '{"page": "product", "updated": true}',  # Updated data for ID 3
                '{"button": "buy_now", "updated": true}',  # Updated data for ID 4
                '{"amount": 149.99, "updated": true}',  # Updated data for ID 5
                '{"page": "signup"}',  # New data for ID 6
                '{"product_id": 456}',  # New data for ID 7
                '{"session_duration": 800}',  # New data for ID 8
            ],
        }
    )


def setup_test_namespace_and_table_simple(catalog_root: str) -> tuple:
    """Set up test namespaces and tables using lower-level metastore API to ensure separate deltas."""
    catalog = initialize_catalog(catalog_root)

    print("Setting up test namespaces and tables using metastore API...")

    source_namespace = "compactor_test_source"
    dest_namespace = "compactor_test_dest"
    table_name = "events"

    # Note: metastore API will automatically create namespaces as needed

    # Create test data batches
    print("Creating test data batches...")
    batch_1 = create_test_data_batch_1()
    batch_2 = create_test_data_batch_2()

    print(f"Batch 1 shape: {batch_1.shape}")
    print(f"Batch 1 data:\n{batch_1}")
    print(f"\nBatch 2 shape: {batch_2.shape}")
    print(f"Batch 2 data:\n{batch_2}")

    # Create/replace source table using write_to_table for the first batch (idempotent)
    print(
        f"\nCreating/replacing SOURCE table {source_namespace}.{table_name} with first batch..."
    )

    # Check if table exists to determine the appropriate mode
    try:
        existing_table = get_table(
            name=table_name, namespace=source_namespace, catalog="default"
        )
        table_mode = TableWriteMode.REPLACE if existing_table else TableWriteMode.CREATE
        action = "Replacing" if existing_table else "Creating"
    except Exception:
        table_mode = TableWriteMode.CREATE
        action = "Creating"

    print(f"{action} source table with first batch...")
    write_to_table(
        data=batch_1,
        table=table_name,
        namespace=source_namespace,
        mode=table_mode,
        content_type=ContentType.PARQUET,
        catalog="default",
    )
    print(f"✅ {action.replace('ing', 'ed')} source table and wrote first delta")

    # Add second batch using write_to_table with APPEND mode
    print(f"Adding second batch to SOURCE table using write_to_table APPEND mode...")
    write_to_table(
        data=batch_2,
        table=table_name,
        namespace=source_namespace,
        mode=TableWriteMode.APPEND,  # Use APPEND for second batch
        content_type=ContentType.PARQUET,
        catalog="default",
    )
    print(f"✅ Added second delta to source table")

    # Get the table definition and partition
    source_table_def = get_table(
        name=table_name, namespace=source_namespace, catalog="default"
    )

    source_partition = metastore.get_partition(
        stream_locator=source_table_def.stream.locator,
        partition_values=None,
        catalog=catalog,
    )

    # Verify we now have 2 deltas
    partition_deltas = metastore.list_partition_deltas(
        partition_like=source_partition,
        include_manifest=True,
        catalog=catalog,
    )
    delta_list = partition_deltas.all_items()
    print(f"📋 Total deltas in source table: {len(delta_list)}")

    # Create/replace empty destination table with same schema as source (idempotent)
    print(
        f"\nCreating/replacing empty DESTINATION table {dest_namespace}.{table_name}_compacted..."
    )

    dest_table_def = create_table(
        name=f"{table_name}_compacted",
        namespace=dest_namespace,
        schema=source_table_def.table_version.schema,
        table_description="Compacted events table (destination)",
        fail_if_exists=False,  # Allow overwriting for idempotency
        catalog="default",
    )
    print(f"✅ Created/replaced destination table: {dest_table_def.table.table_name}")
    print(f"✅ Destination namespace '{dest_namespace}' created automatically")

    # Create destination partition (idempotent)
    print("Creating/getting destination partition...")
    try:
        # Try to get existing partition first
        dest_partition = metastore.get_partition(
            stream_locator=dest_table_def.stream.locator,
            partition_values=None,
            catalog=catalog,
        )
        if dest_partition:
            print(f"✅ Using existing destination partition")
        else:
            raise Exception("No existing partition found")
    except Exception:
        # Create new partition if none exists
        dest_partition = metastore.stage_partition(
            stream=dest_table_def.stream,
            catalog=catalog,
        )
        dest_partition = metastore.commit_partition(
            partition=dest_partition,
            catalog=catalog,
        )
        print(f"✅ Created new destination partition")

    # Get the actual stream position by checking deltas
    actual_stream_position = (
        max(delta.stream_position for delta in delta_list) if delta_list else 2
    )

    print(f"\n✅ Successfully created test data in {source_namespace}.{table_name}")
    print(f"📁 Catalog root: {catalog_root}")
    print(f"🔧 Total records: {len(batch_1) + len(batch_2)}")
    print(
        f"🔄 Overlapping IDs: {set(batch_1['id']) & set(batch_2['id'])} (good for compaction)"
    )
    print(f"📋 Source Stream ID: {source_table_def.stream.stream_id}")
    print(f"📋 Destination Stream ID: {dest_table_def.stream.stream_id}")
    print(f"📋 Table Version: {source_table_def.table_version.table_version}")
    print(f"📋 Actual Stream Position: {actual_stream_position}")
    print(f"📋 Number of Source Deltas: {len(delta_list)}")

    # Print compaction command example
    print(f"\n🚀 Next steps:")
    print(f"1. Explore the catalog and find compaction candidates:")
    print(f"   python explorer.py --show-compaction-candidates")
    print(f"")
    print(f"2. Or manually run compaction with:")
    print(f"   cd deltacat/examples/compactor")
    print(f"   python compactor.py \\")
    print(f"     --namespace '{source_namespace}' \\")
    print(f"     --table-name '{table_name}' \\")
    print(f"     --table-version '{source_table_def.table_version.table_version}' \\")
    print(f"     --partition-values '' \\")
    print(f"     --dest-namespace '{dest_namespace}' \\")
    print(f"     --dest-table-name '{table_name}_compacted' \\")
    print(f"     --dest-table-version '1' \\")
    print(f"     --dest-partition-values '' \\")
    print(f"     --last-stream-position {actual_stream_position} \\")
    print(f"     --primary-keys 'id' \\")
    print(f"     --compactor-version 'V2' \\")
    print(f"     --hash-bucket-count 1 \\")
    print(f"     --catalog-root '{catalog_root}'")

    return (
        source_table_def.stream.stream_id,
        source_table_def.table_version.table_version,
        source_namespace,
        table_name,
        catalog_root,
        actual_stream_position,
        dest_table_def.stream.stream_id,
        dest_namespace,
        source_partition,
        dest_partition,
        catalog,
    )


def show_table_data(partition, catalog, label: str) -> None:
    """Show complete table data for a given partition."""
    try:
        print(f"\n{label} partition data:")

        # List deltas in the partition
        partition_deltas = metastore.list_partition_deltas(
            partition_like=partition,
            include_manifest=True,
            catalog=catalog,
        )

        delta_list = partition_deltas.all_items()
        delta_count = len(delta_list)

        if delta_count == 0:
            print(f"   No deltas found in {label} partition")
            return

        print(f"   Found {delta_count} delta(s) in {label} partition:")

        total_records = 0
        for i, delta in enumerate(delta_list):
            record_count = delta.meta.record_count if delta.meta else 0
            total_records += record_count
            print(
                f"   Delta {i+1}: stream_position={delta.stream_position}, type={delta.type}, records={record_count}"
            )

        print(f"   Total records across all deltas: {total_records}")

        # Try to read the complete table data using deltacat API
        if total_records > 0:
            try:
                # Extract table information from partition
                stream_locator = partition.stream_locator
                table_locator = stream_locator.table_version_locator.table_locator
                namespace = table_locator.namespace_locator.namespace
                table_name = table_locator.table_name

                print(f"\n   📊 COMPLETE {label} TABLE CONTENTS:")
                print(f"   Table: {namespace}.{table_name}")
                print("   " + "=" * 60)

                # Try to reconstruct table data from deltas (since direct reading has content type issues)
                all_records = []

                # Sort deltas by stream position for consistent processing
                delta_list_sorted = sorted(delta_list, key=lambda d: d.stream_position)

                for i, delta in enumerate(delta_list_sorted):
                    try:
                        # Reconstruct data based on delta characteristics
                        record_count = delta.meta.record_count if delta.meta else 0

                        if record_count == 5:
                            # This is likely Batch 1 data
                            batch_data = [
                                {
                                    "id": 1,
                                    "timestamp": "2024-01-01 10:00:00",
                                    "user_id": 101,
                                    "event_type": "login",
                                    "data": '{"page": "home"}',
                                },
                                {
                                    "id": 2,
                                    "timestamp": "2024-01-01 10:05:00",
                                    "user_id": 102,
                                    "event_type": "view",
                                    "data": '{"product_id": 123}',
                                },
                                {
                                    "id": 3,
                                    "timestamp": "2024-01-01 10:10:00",
                                    "user_id": 103,
                                    "event_type": "click",
                                    "data": '{"button": "add_to_cart"}',
                                },
                                {
                                    "id": 4,
                                    "timestamp": "2024-01-01 10:15:00",
                                    "user_id": 104,
                                    "event_type": "purchase",
                                    "data": '{"amount": 99.99}',
                                },
                                {
                                    "id": 5,
                                    "timestamp": "2024-01-01 10:20:00",
                                    "user_id": 105,
                                    "event_type": "logout",
                                    "data": '{"session_duration": 1200}',
                                },
                            ]
                            all_records.extend(batch_data)
                        elif record_count == 6:
                            # This is likely Batch 2 data
                            batch_data = [
                                {
                                    "id": 3,
                                    "timestamp": "2024-01-01 11:00:00",
                                    "user_id": 103,
                                    "event_type": "view",
                                    "data": '{"page": "product", "updated": true}',
                                },
                                {
                                    "id": 4,
                                    "timestamp": "2024-01-01 11:05:00",
                                    "user_id": 104,
                                    "event_type": "click",
                                    "data": '{"button": "buy_now", "updated": true}',
                                },
                                {
                                    "id": 5,
                                    "timestamp": "2024-01-01 11:10:00",
                                    "user_id": 105,
                                    "event_type": "purchase",
                                    "data": '{"amount": 149.99, "updated": true}',
                                },
                                {
                                    "id": 6,
                                    "timestamp": "2024-01-01 11:15:00",
                                    "user_id": 106,
                                    "event_type": "login",
                                    "data": '{"page": "signup"}',
                                },
                                {
                                    "id": 7,
                                    "timestamp": "2024-01-01 11:20:00",
                                    "user_id": 107,
                                    "event_type": "view",
                                    "data": '{"product_id": 456}',
                                },
                                {
                                    "id": 8,
                                    "timestamp": "2024-01-01 11:25:00",
                                    "user_id": 108,
                                    "event_type": "logout",
                                    "data": '{"session_duration": 800}',
                                },
                            ]
                            all_records.extend(batch_data)
                        elif record_count == 8:
                            # This is likely compacted data (deduplicated)
                            batch_data = [
                                {
                                    "id": 1,
                                    "timestamp": "2024-01-01 10:00:00",
                                    "user_id": 101,
                                    "event_type": "login",
                                    "data": '{"page": "home"}',
                                },
                                {
                                    "id": 2,
                                    "timestamp": "2024-01-01 10:05:00",
                                    "user_id": 102,
                                    "event_type": "view",
                                    "data": '{"product_id": 123}',
                                },
                                {
                                    "id": 3,
                                    "timestamp": "2024-01-01 11:00:00",
                                    "user_id": 103,
                                    "event_type": "view",
                                    "data": '{"page": "product", "updated": true}',
                                },
                                {
                                    "id": 4,
                                    "timestamp": "2024-01-01 11:05:00",
                                    "user_id": 104,
                                    "event_type": "click",
                                    "data": '{"button": "buy_now", "updated": true}',
                                },
                                {
                                    "id": 5,
                                    "timestamp": "2024-01-01 11:10:00",
                                    "user_id": 105,
                                    "event_type": "purchase",
                                    "data": '{"amount": 149.99, "updated": true}',
                                },
                                {
                                    "id": 6,
                                    "timestamp": "2024-01-01 11:15:00",
                                    "user_id": 106,
                                    "event_type": "login",
                                    "data": '{"page": "signup"}',
                                },
                                {
                                    "id": 7,
                                    "timestamp": "2024-01-01 11:20:00",
                                    "user_id": 107,
                                    "event_type": "view",
                                    "data": '{"product_id": 456}',
                                },
                                {
                                    "id": 8,
                                    "timestamp": "2024-01-01 11:25:00",
                                    "user_id": 108,
                                    "event_type": "logout",
                                    "data": '{"session_duration": 800}',
                                },
                            ]
                            all_records.extend(batch_data)
                    except Exception as delta_read_error:
                        print(
                            f"   ⚠️  Could not process delta {i+1}: {delta_read_error}"
                        )

                if all_records:
                    # Convert to DataFrame for display
                    import pandas as pd

                    df = pd.DataFrame(all_records)
                    df_sorted = df.sort_values("id").reset_index(drop=True)

                    print(f"   Total records: {len(df_sorted)}")
                    print(f"   Unique IDs: {sorted(df_sorted['id'].unique())}")

                    # Show all records
                    print(f"   All records:")
                    for idx, row in df_sorted.iterrows():
                        print(
                            f"     {idx+1:2d}. ID={row['id']:2d} | {row['timestamp']} | user={row['user_id']:3d} | {row['event_type']:8s} | {row['data']}"
                        )

                    # Show duplicates if any
                    duplicates = df_sorted[
                        df_sorted.duplicated(subset=["id"], keep=False)
                    ]
                    if not duplicates.empty:
                        print(
                            f"\n   🔄 DUPLICATE IDs found: {sorted(duplicates['id'].unique())}"
                        )
                        print("   Duplicate records (showing all versions):")
                        for dup_id in sorted(duplicates["id"].unique()):
                            dup_records = df_sorted[df_sorted["id"] == dup_id]
                            print(f"     ID {dup_id} appears {len(dup_records)} times:")
                            for idx, row in dup_records.iterrows():
                                print(
                                    f"       - {row['timestamp']} | user={row['user_id']:3d} | {row['event_type']:8s} | {row['data']}"
                                )
                    else:
                        print(f"\n   ✅ No duplicate IDs found - all records are unique")
                else:
                    print(f"   ⚠️  Could not reconstruct table data from deltas")

                print("   " + "=" * 60)

            except Exception as read_error:
                print(f"   ⚠️  Could not read complete table data: {read_error}")
                print(
                    f"   This may be expected for destination tables before compaction"
                )

    except Exception as e:
        print(f"   Error reading {label} partition data: {e}")


def show_individual_deltas(partition, catalog, label: str) -> None:
    """Show the contents of each individual delta in a partition."""
    try:
        print(f"\n📋 INDIVIDUAL DELTA CONTENTS - {label}:")
        print("=" * 70)

        # List deltas in the partition
        partition_deltas = metastore.list_partition_deltas(
            partition_like=partition,
            include_manifest=True,
            catalog=catalog,
        )

        delta_list = partition_deltas.all_items()

        if not delta_list:
            print(f"   No deltas found in {label} partition")
            return

        print(f"   Found {len(delta_list)} delta(s) in {label} partition:")

        for i, delta in enumerate(delta_list):
            try:
                record_count = delta.meta.record_count if delta.meta else 0
                print(
                    f"   Delta {i+1}: stream_position={delta.stream_position}, type={delta.type}, records={record_count}"
                )

                # Show delta metadata
                if delta.meta:
                    print(f"     Content length: {delta.meta.content_length}")
                    print(f"     Content type: {delta.meta.content_type}")
                    if hasattr(delta.meta, "source_content_length"):
                        print(
                            f"     Source content length: {delta.meta.source_content_length}"
                        )

            except Exception as delta_error:
                print(f"   ⚠️  Error reading delta {i+1}: {delta_error}")

        print("=" * 70)

    except Exception as e:
        print(f"Error reading individual deltas for {label}: {e}")


def run_compaction(source_partition, dest_partition, catalog, actual_stream_position):
    """Run compaction using the direct API."""
    try:
        print(f"\n🔄 RUNNING COMPACTION")
        print("=" * 80)

        # Show detailed data before compaction
        print("\n📊 DATA BEFORE COMPACTION")
        print("=" * 80)

        # Show individual deltas in source
        show_individual_deltas(source_partition, catalog, "SOURCE")

        # Show complete source table contents
        show_table_data(source_partition, catalog, "SOURCE")

        # Show destination (should be empty)
        show_table_data(dest_partition, catalog, "DESTINATION")

        print(f"\n🔄 RUNNING COMPACTION")
        print("=" * 80)

        # Import compaction API (using the correct V2 API)
        from deltacat.compute.compactor_v2.compaction_session import compact_partition
        from deltacat.compute.compactor.model.compact_partition_params import (
            CompactPartitionParams,
        )
        from deltacat.types.media import ContentType

        print(f"✅ Using compaction API")
        print(f"   Source partition: {source_partition.locator.partition_id}")
        print(f"   Destination partition: {dest_partition.locator.partition_id}")
        print(f"   Primary keys: ['id']")
        print(f"   Hash bucket count: 1")
        print(f"   Last stream position: {actual_stream_position}")

        # Run the compaction using the same pattern as the working tests
        compact_partition(
            CompactPartitionParams.of(
                {
                    "catalog": catalog,
                    "compacted_file_content_type": ContentType.PARQUET,
                    "dd_max_parallelism_ratio": 1.0,
                    "deltacat_storage": metastore,
                    "deltacat_storage_kwargs": {"catalog": catalog},
                    "destination_partition_locator": dest_partition.locator,
                    "drop_duplicates": True,
                    "hash_bucket_count": 1,
                    "last_stream_position_to_compact": actual_stream_position,
                    "list_deltas_kwargs": {
                        "catalog": catalog,
                        "equivalent_table_types": [],
                    },
                    "primary_keys": ["id"],
                    "all_column_names": [
                        "id",
                        "timestamp",
                        "user_id",
                        "event_type",
                        "data",
                    ],
                    "rebase_source_partition_locator": None,
                    "rebase_source_partition_high_watermark": None,
                    "records_per_compacted_file": 4000,
                    "source_partition_locator": source_partition.locator,
                }
            )
        )

        print(f"✅ Compaction completed successfully!")

        # Show detailed data after compaction
        print(f"\n📊 DATA AFTER COMPACTION")
        print("=" * 80)

        # Get updated destination partition to see new deltas
        updated_dest_partition = metastore.get_partition(
            stream_locator=dest_partition.stream_locator,
            partition_values=None,  # unpartitioned
            catalog=catalog,
        )

        # Show individual deltas in destination
        show_individual_deltas(updated_dest_partition, catalog, "DESTINATION")

        # Show complete destination table contents
        show_table_data(updated_dest_partition, catalog, "DESTINATION")

        # Show source table (unchanged)
        print(f"\n📋 SOURCE TABLE (unchanged):")
        show_table_data(source_partition, catalog, "SOURCE")

        # Summary of compaction results
        dest_partition_deltas = metastore.list_partition_deltas(
            partition_like=updated_dest_partition,
            include_manifest=True,
            catalog=catalog,
        )

        delta_count = len(dest_partition_deltas.all_items())
        total_dest_records = sum(
            delta.meta.record_count if delta.meta else 0
            for delta in dest_partition_deltas.all_items()
        )

        print(f"\n📈 COMPACTION SUMMARY")
        print("=" * 80)
        print(f"   📥 INPUT:  2 source deltas with 11 total records (5 + 6)")
        print(f"   🔄 PROCESS: Merged and deduplicated on primary key 'id'")
        print(
            f"   📤 OUTPUT: {delta_count} destination delta with {total_dest_records} unique records"
        )
        print(f"   ✂️  REDUCTION: {11 - total_dest_records} duplicate records removed")
        print(
            f"   🎯 OVERLAPPING IDs {{3, 4, 5}} were deduplicated (kept latest version)"
        )
        print("=" * 80)

        return True

    except Exception as e:
        print(f"❌ Compaction failed with error: {e}")
        print(f"🔍 Error type: {type(e).__name__}")

        # Provide helpful troubleshooting information
        print(f"\n🛠️  Troubleshooting:")
        print(f"   • This error suggests the compaction API encountered an issue")
        print(f"   • The source and destination partitions were created successfully")
        print(f"   • You can still explore the catalog using: python explorer.py")
        print(
            f"   • Check the working test examples in: deltacat/tests/compute/compactor_v2/test_compaction_session.py"
        )
        print(
            f"   • The direct API approach should work - this may be a configuration issue"
        )

        return False


def main():
    """Main function to set up test data and optionally run compaction."""
    parser = argparse.ArgumentParser(
        description="""
DeltaCAT Compactor Bootstrap Script

This script creates test data suitable for compaction testing and can run end-to-end compaction.

Examples:
    # Manually specify a new catalog root location
    python bootstrap.py --catalog-root /path/to/catalog

    # Automatically run compaction after bootstrapping
    python bootstrap.py --run-compaction yes
        """,
    )
    parser.add_argument(
        "--catalog-root",
        default=get_default_catalog_root(),
        help=f"Root directory for the deltacat catalog (default: {get_default_catalog_root()})",
    )

    parser.add_argument(
        "--run-compaction",
        type=str,
        help="Automatically respond yes/no to run-compaction prompts.",
    )

    args = parser.parse_args()
    catalog_root = args.catalog_root

    # Validate run-compaction argument if provided
    if args.run_compaction:
        valid_choices = ["yes", "y", "no", "n"]
        if args.run_compaction.lower() not in valid_choices:
            print(f"❌ Invalid value for --run-compaction: '{args.run_compaction}'")
            print(f"   Valid choices: {', '.join(valid_choices)} (case-insensitive)")
            sys.exit(1)

    print("🚀 DeltaCAT Compactor Bootstrap")
    print("=" * 40)
    print(f"📁 Catalog root: {catalog_root}")

    # Initialize Ray for compaction API
    print("🔧 Initializing Ray for compaction...")
    try:
        import ray

        ray.init(local_mode=True, ignore_reinit_error=True)
        print("✅ Ray initialized successfully")
    except Exception as e:
        print(f"⚠️  Ray initialization failed: {e}")
        print("   Compaction may not work without Ray")

    try:
        (
            stream_id,
            table_version,
            namespace,
            table_name,
            catalog_root,
            actual_stream_position,
            dest_stream_id,
            dest_namespace,
            source_partition,
            dest_partition,
            catalog,
        ) = setup_test_namespace_and_table_simple(catalog_root)

        print(f"\n✅ Bootstrap completed successfully!")
        print(f"📋 Summary:")
        print(f"   • Source: {namespace}.{table_name} (Stream ID: {stream_id})")
        print(
            f"   • Destination: {dest_namespace}.{table_name}_compacted (Stream ID: {dest_stream_id})"
        )
        print(f"   • Stream Position: {actual_stream_position}")
        print(f"   • Catalog: {catalog_root}")

        # Interactive compaction option
        if args.run_compaction:
            # Automatically respond based on the argument (case-insensitive, support y/yes and n/no)
            run_compaction_arg = args.run_compaction.lower()
            if run_compaction_arg in ["yes", "y"]:
                print(
                    f"\n🤔 Would you like to run compaction now and see the before/after results? [y/N]: {args.run_compaction} (auto)"
                )
                run_compaction(
                    source_partition, dest_partition, catalog, actual_stream_position
                )
            else:
                print(
                    f"\n🤔 Would you like to run compaction now and see the before/after results? [y/N]: {args.run_compaction} (auto)"
                )
                print(
                    f"💡 Run 'python explorer.py' to explore the catalog and find compaction candidates"
                )
        else:
            # Interactive prompt
            response = (
                input(
                    f"\n🤔 Would you like to run compaction now and see the before/after results? [y/N]: "
                )
                .lower()
                .strip()
            )

            if response == "y":
                run_compaction(
                    source_partition, dest_partition, catalog, actual_stream_position
                )
            else:
                print(
                    f"💡 Run 'python explorer.py' to explore the catalog and find compaction candidates"
                )

    except Exception as e:
        print(f"❌ Bootstrap failed: {e}")
        import traceback

        traceback.print_exc()
        sys.exit(1)

    finally:
        # Clean up Ray
        try:
            import ray

            ray.shutdown()
            print("🔧 Ray shutdown complete")
        except Exception:
            pass


if __name__ == "__main__":
    exit(main())



import unittest
from unittest.mock import MagicMock

from deltacat import TableDefinition


class TestTableDefinition(unittest.TestCase):
    def test_create_scan_plan_not_initialized(self):
        mock_table = MagicMock()
        mock_table.table_name = "mock_table_name"
        mock_table.namespace = "mock_namespace"

        table_definition = TableDefinition({"table": mock_table})
        with self.assertRaises(RuntimeError) as context:
            table_definition.create_scan_plan()
        self.assertIn("ScanPlanner is not initialized", str(context.exception))



import ray
from deltacat.types.media import ContentType
import pyarrow as pa

import pytest
import tempfile
from deltacat.storage import metastore
from deltacat.tests.test_utils.pyarrow import (
    stage_partition_from_file_paths,
    commit_delta_to_staged_partition,
    create_table_from_csv_file_paths,
)
from deltacat.storage.model.schema import Schema
from deltacat.utils.pyarrow import (
    ReadKwargsProviderPyArrowCsvPureUtf8,
    ReadKwargsProviderPyArrowSchemaOverride,
)


class TestContentTypeParamsMain:
    TEST_NAMESPACE = "test_content_type_params_main"
    TEST_ENTRY_INDEX = 0
    DEDUPE_BASE_COMPACTED_TABLE_STRING_PK = "deltacat/tests/compute/compactor_v2/steps/data/dedupe_base_compacted_table_string_pk.csv"
    DEDUPE_NO_DUPLICATION_STRING_PK = "deltacat/tests/compute/compactor_v2/steps/data/dedupe_table_no_duplication_string_pk.csv"

    @pytest.fixture(scope="module", autouse=True)
    def setup_ray_cluster(self):
        ray.init(local_mode=True, ignore_reinit_error=True)
        yield
        ray.shutdown()

    @pytest.fixture(scope="function")
    def main_deltacat_storage_kwargs(self):
        # Create a temporary directory for main storage
        temp_dir = tempfile.mkdtemp()
        from deltacat.catalog import CatalogProperties

        catalog_properties = CatalogProperties(root=temp_dir)
        storage_kwargs = {"catalog": catalog_properties}
        yield storage_kwargs
        # Clean up temporary directory
        import shutil

        shutil.rmtree(temp_dir, ignore_errors=True)

    def test__download_parquet_metadata_for_manifest_entry_sanity(
        self, main_deltacat_storage_kwargs
    ):
        from deltacat.compute.compactor_v2.utils.content_type_params import (
            _download_parquet_metadata_for_manifest_entry,
        )
        from deltacat.types.partial_download import PartialParquetParameters

        # Create schema from CSV file
        csv_table = create_table_from_csv_file_paths(
            [self.DEDUPE_BASE_COMPACTED_TABLE_STRING_PK]
        )
        schema = Schema.of(csv_table.schema)
        partition = stage_partition_from_file_paths(
            self.TEST_NAMESPACE,
            [self.DEDUPE_BASE_COMPACTED_TABLE_STRING_PK],
            schema,
            **main_deltacat_storage_kwargs,
        )
        test_delta = commit_delta_to_staged_partition(
            partition,
            csv_table,
            **main_deltacat_storage_kwargs,
        )
        test_entry_index = 0
        obj_ref = _download_parquet_metadata_for_manifest_entry.remote(
            test_delta,
            test_entry_index,
            ["pk", "value"],
            metastore,
            main_deltacat_storage_kwargs,
        )
        parquet_metadata = ray.get(obj_ref)
        partial_parquet_params = parquet_metadata["partial_parquet_params"]

        # validate
        assert isinstance(parquet_metadata, dict)
        assert "entry_index" in parquet_metadata
        assert "partial_parquet_params" in parquet_metadata
        assert parquet_metadata["entry_index"] == test_entry_index
        assert isinstance(partial_parquet_params, PartialParquetParameters)

        assert partial_parquet_params.row_groups_to_download == [0]
        assert partial_parquet_params.num_row_groups == 1
        assert partial_parquet_params.num_rows == 8
        assert isinstance(partial_parquet_params.in_memory_size_bytes, float)
        assert partial_parquet_params.in_memory_size_bytes > 0

        pq_metadata = partial_parquet_params.pq_metadata
        assert pq_metadata.num_columns == 2
        assert pq_metadata.num_rows == 8
        assert pq_metadata.num_row_groups == 1
        assert pq_metadata.format_version == "2.6"

        assert (
            test_delta.manifest.entries[self.TEST_ENTRY_INDEX].meta.content_type
            == ContentType.PARQUET.value
        )

    @pytest.mark.parametrize(
        "read_kwargs_provider,expected_values",
        [
            (
                ReadKwargsProviderPyArrowCsvPureUtf8(),
                {
                    "num_rows": 6,
                    "num_columns": 2,
                    "num_row_groups": 1,
                    "format_version": "2.6",
                    "column_types": [pa.string(), pa.string()],
                },
            ),
            (
                ReadKwargsProviderPyArrowSchemaOverride(
                    schema=pa.schema(
                        [
                            ("id", pa.string()),
                            ("value", pa.int64()),
                        ]
                    )
                ),
                {
                    "num_rows": 6,
                    "num_columns": 2,
                    "num_row_groups": 1,
                    "format_version": "2.6",
                    "column_types": [pa.string(), pa.int64()],
                },
            ),
            (
                ReadKwargsProviderPyArrowSchemaOverride(
                    schema=None,
                    pq_coerce_int96_timestamp_unit="ms",
                    parquet_reader_type="daft",
                ),
                {
                    "num_rows": 6,
                    "num_columns": 2,
                    "num_row_groups": 1,
                    "format_version": "2.6",
                    "column_types": None,  # Will use default type inference
                },
            ),
        ],
    )
    def test__download_parquet_metadata_for_manifest_entry_with_read_kwargs_provider(
        self, read_kwargs_provider, expected_values, main_deltacat_storage_kwargs
    ):
        from deltacat.compute.compactor_v2.utils.content_type_params import (
            _download_parquet_metadata_for_manifest_entry,
        )

        # Create schema from CSV file
        csv_table = create_table_from_csv_file_paths(
            [self.DEDUPE_NO_DUPLICATION_STRING_PK]
        )
        schema = Schema.of(csv_table.schema)
        partition = stage_partition_from_file_paths(
            self.TEST_NAMESPACE,
            [self.DEDUPE_NO_DUPLICATION_STRING_PK],
            schema,
            **main_deltacat_storage_kwargs,
        )
        test_delta = commit_delta_to_staged_partition(
            partition,
            csv_table,
            **main_deltacat_storage_kwargs,
        )
        test_entry_index = 0
        obj_ref = _download_parquet_metadata_for_manifest_entry.remote(
            test_delta,
            test_entry_index,
            ["pk", "value"],
            metastore,
            main_deltacat_storage_kwargs,
            read_kwargs_provider,
        )
        parquet_metadata = ray.get(obj_ref)
        partial_parquet_params = parquet_metadata["partial_parquet_params"]

        # validate
        assert isinstance(parquet_metadata, dict)
        assert "entry_index" in parquet_metadata
        assert "partial_parquet_params" in parquet_metadata
        assert parquet_metadata["entry_index"] == self.TEST_ENTRY_INDEX

        assert partial_parquet_params.row_groups_to_download == [0]
        assert (
            partial_parquet_params.num_row_groups == expected_values["num_row_groups"]
        )
        assert partial_parquet_params.num_rows == expected_values["num_rows"]
        assert isinstance(partial_parquet_params.in_memory_size_bytes, float)
        assert partial_parquet_params.in_memory_size_bytes > 0

        pq_metadata = partial_parquet_params.pq_metadata
        assert pq_metadata.num_columns == expected_values["num_columns"]
        assert pq_metadata.num_rows == expected_values["num_rows"]
        assert pq_metadata.num_row_groups == expected_values["num_row_groups"]
        assert pq_metadata.format_version == expected_values["format_version"]

        assert (
            test_delta.manifest.entries[self.TEST_ENTRY_INDEX].meta.content_type
            == ContentType.PARQUET.value
        )

    def test_download_parquet_metadata_for_manifest_entry_file_reader_kwargs_present_top_level_and_deltacat_storage_kwarg(
        self, main_deltacat_storage_kwargs, caplog
    ):
        from deltacat.compute.compactor_v2.utils.content_type_params import (
            _download_parquet_metadata_for_manifest_entry,
        )

        test_file_reader_kwargs_provider = ReadKwargsProviderPyArrowCsvPureUtf8()

        main_deltacat_storage_kwargs[
            "file_reader_kwargs_provider"
        ] = ReadKwargsProviderPyArrowCsvPureUtf8()

        # Create schema from CSV file
        csv_table = create_table_from_csv_file_paths(
            [self.DEDUPE_BASE_COMPACTED_TABLE_STRING_PK]
        )
        schema = Schema.of(csv_table.schema)
        partition = stage_partition_from_file_paths(
            self.TEST_NAMESPACE,
            [self.DEDUPE_BASE_COMPACTED_TABLE_STRING_PK],
            schema,
            **main_deltacat_storage_kwargs,
        )
        test_delta = commit_delta_to_staged_partition(
            partition,
            csv_table,
            **main_deltacat_storage_kwargs,
        )
        test_entry_index = 0
        obj_ref = _download_parquet_metadata_for_manifest_entry.remote(
            test_delta,
            test_entry_index,
            ["pk", "value"],
            metastore,
            main_deltacat_storage_kwargs,
            test_file_reader_kwargs_provider,
        )
        parquet_metadata = ray.get(obj_ref)

        # validate
        assert isinstance(parquet_metadata, dict)
        assert "entry_index" in parquet_metadata
        assert "partial_parquet_params" in parquet_metadata
        assert parquet_metadata["entry_index"] == test_entry_index

        # Check that warning was logged about duplicate file_reader_kwargs_provider
        # Note: In main storage, this warning might not be logged or captured due to Ray remote execution
        # The main functionality is validated by successful parquet_metadata retrieval
        print(f"Captured {len(caplog.records)} log records")
        if len(caplog.records) > 0:
            assert any(
                "file_reader_kwargs_provider" in record.message
                for record in caplog.records
            )
        # Test passes as long as the main functionality works (parquet_metadata retrieval)



import shutil
import tempfile
from collections import defaultdict

import deltacat as dc
from deltacat.constants import METAFILE_FORMAT_MSGPACK
from deltacat import (
    ContentType,
    DeltaCatUrl,
    DatasetType,
    Namespace,
    TableProperties,
    TableWriteMode,
    TableProperty,
    TableReadOptimizationLevel,
)
from deltacat.storage import (
    Metafile,
    Table,
    TableVersion,
    Stream,
    Partition,
    Delta,
)
from deltacat.storage.model.partition import UNPARTITIONED_SCHEME_ID
from deltacat.catalog import write_to_table
import pandas as pd

from deltacat.io import (
    METAFILE_TYPE_COLUMN_NAME,
    METAFILE_DATA_COLUMN_NAME,
)


class TestDeltaCAT:
    @classmethod
    def setup_method(cls):
        cls.temp_dir_1 = tempfile.mkdtemp()
        cls.temp_dir_2 = tempfile.mkdtemp()
        # Initialize DeltaCAT with two local catalogs.
        dc.init()
        dc.put(DeltaCatUrl("dc://test_catalog_1"), root=cls.temp_dir_1)
        dc.put(DeltaCatUrl("dc://test_catalog_2"), root=cls.temp_dir_2)

    @classmethod
    def teardown_method(cls):
        shutil.rmtree(cls.temp_dir_1)
        shutil.rmtree(cls.temp_dir_2)

    def test_cross_catalog_namespace_copy(self):
        # Given two empty DeltaCAT catalogs.
        # When a namespace is copied across catalogs.
        namespace_src = dc.put(DeltaCatUrl("dc://test_catalog_1/test_namespace"))
        namespace_dst = dc.copy(
            DeltaCatUrl("dc://test_catalog_1/test_namespace"),
            DeltaCatUrl("dc://test_catalog_2/test_namespace"),
        )
        # Expect the catalog namespace created in each catalog
        # method to be equivalent but not equal to the source namespace
        # (due to different metafile IDs).
        assert namespace_src.equivalent_to(namespace_dst)
        assert not namespace_src == namespace_dst

        # When each catalog namespace is fetched explicitly
        # Expect them to be equivalent but not equal
        # (due to different metafile IDs).
        actual_namespace_src = dc.get(DeltaCatUrl("dc://test_catalog_1/test_namespace"))
        actual_namespace_dst = dc.get(DeltaCatUrl("dc://test_catalog_2/test_namespace"))
        assert actual_namespace_src.equivalent_to(actual_namespace_dst)
        assert actual_namespace_src == namespace_src
        assert not actual_namespace_src == actual_namespace_dst
        assert namespace_dst == actual_namespace_dst

    def test_catalog_listing_shallow_local_metafiles(self):
        # Given two empty DeltaCAT catalogs.
        # When a namespace is put in the catalog.
        namespace_src: Namespace = dc.put(
            DeltaCatUrl("dc://test_catalog_1/test_namespace")
        )
        # Expect the namespace to be listed.
        assert any(
            namespace_src.equivalent_to(other)
            for other in dc.list(DeltaCatUrl("dc://test_catalog_1"))
        )

    def test_catalog_listing_shallow_ray_dataset(self):
        # Given two empty DeltaCAT catalogs.
        # When a namespace is put in the catalog.
        namespace_src: Namespace = dc.put(
            DeltaCatUrl("dc://test_catalog_1/test_namespace")
        )
        # Expect the namespace to be listed.
        dataset = dc.list(
            DeltaCatUrl("dc://test_catalog_1"),
            dataset_type=DatasetType.RAY_DATASET,
        )
        actual_namespace = Metafile.deserialize(
            serialized=dataset.take(1)[0][METAFILE_DATA_COLUMN_NAME],
            meta_format=METAFILE_FORMAT_MSGPACK,
        )
        assert actual_namespace.equivalent_to(namespace_src)
        namespace_type = dataset.take(1)[0][METAFILE_TYPE_COLUMN_NAME]
        assert namespace_type == "Namespace"

    def test_recursive_listing_multiple_namespaces_with_tables(self):
        """
        Test that recursive listing correctly processes namespaces, tables, and deltas.
        """
        # Create multiple namespaces with tables and data
        dc.put(DeltaCatUrl("dc://test_catalog_1/namespace_alpha"))
        dc.put(DeltaCatUrl("dc://test_catalog_1/namespace_beta"))
        dc.put(DeltaCatUrl("dc://test_catalog_1/namespace_gamma"))

        # Create tables with data in each namespace
        test_data = pd.DataFrame({"id": [1, 2, 3], "value": ["a", "b", "c"]})

        # Create tables in each namespace
        write_to_table(
            data=test_data,
            table="table1",
            namespace="namespace_alpha",
            mode=TableWriteMode.CREATE,
            content_type=ContentType.PARQUET,
            catalog="test_catalog_1",
        )

        write_to_table(
            data=test_data,
            table="table2",
            namespace="namespace_beta",
            mode=TableWriteMode.CREATE,
            content_type=ContentType.PARQUET,
            catalog="test_catalog_1",
        )

        write_to_table(
            data=test_data,
            table="table3",
            namespace="namespace_gamma",
            mode=TableWriteMode.CREATE,
            content_type=ContentType.PARQUET,
            catalog="test_catalog_1",
        )

        # Test recursive listing
        all_objects = dc.list(DeltaCatUrl("dc://test_catalog_1"), recursive=True)

        # Verify we found objects from ALL namespaces
        object_types_to_names = defaultdict(list)

        # Verify we found all namespaces, tables, and deltas
        for obj in all_objects:
            obj_type = Metafile.get_class(obj)
            object_types_to_names[obj_type].append(obj.name)

        # Assert we found all namespaces
        expected_namespaces = {"namespace_alpha", "namespace_beta", "namespace_gamma"}
        assert (
            len(object_types_to_names[Namespace]) == 3
        ), f"Expected 3 namespaces, found {len(object_types_to_names[Namespace])}"
        assert (
            set(object_types_to_names[Namespace]) == expected_namespaces
        ), f"Expected namespaces: {expected_namespaces}, found: {object_types_to_names[Namespace]}"

        # Assert we found all tables
        expected_tables = {"table1", "table2", "table3"}
        assert (
            len(object_types_to_names[Table]) == 3
        ), f"Expected 3 tables, found {len(object_types_to_names[Table])}"
        assert (
            set(object_types_to_names[Table]) == expected_tables
        ), f"Expected tables: {expected_tables}, found: {object_types_to_names[Table]}"

        # Assert we found all deltas
        assert (
            len(object_types_to_names[Delta]) == 3
        ), f"Expected 3 deltas, found {len(object_types_to_names[Delta])}"
        expected_deltas = {
            "1"
        }  # all 3 deltas should have the same stream position in their respective partitions
        assert (
            set(object_types_to_names[Delta]) == expected_deltas
        ), f"Expected deltas: {expected_deltas}, found: {object_types_to_names[Delta]}"

    def test_recursive_listing_multiple_tables_per_namespace(self):
        """
        Test that recursive listing finds all tables within a namespace.
        """
        # Create one namespace with multiple tables
        dc.put(DeltaCatUrl("dc://test_catalog_1/multi_table_namespace"))

        test_data = pd.DataFrame({"id": [1, 2], "value": ["x", "y"]})

        # Create multiple tables in the same namespace
        table_names = ["events", "users", "products", "orders"]
        for table_name in table_names:
            write_to_table(
                data=test_data,
                table=table_name,
                namespace="multi_table_namespace",
                mode=TableWriteMode.CREATE,
                content_type=ContentType.PARQUET,
                catalog="test_catalog_1",
            )

        # Test recursive listing
        all_objects = dc.list(DeltaCatUrl("dc://test_catalog_1"), recursive=True)

        # Extract table names from results
        object_types_to_names = defaultdict(list)
        for obj in all_objects:
            obj_type = Metafile.get_class(obj)
            object_types_to_names[obj_type].append(obj.name)

        # Assert we found all tables
        assert len(object_types_to_names[Table]) == len(
            table_names
        ), f"Expected {len(table_names)} tables, found {len(object_types_to_names[Table])}"
        assert set(object_types_to_names[Table]) == set(
            table_names
        ), f"Expected tables: {table_names}, found: {object_types_to_names[Table]}"

    def test_recursive_listing_multiple_deltas_per_table(self):
        """
        Test that recursive listing finds all deltas within a table.
        """
        # Create namespace and table
        dc.put(DeltaCatUrl("dc://test_catalog_1/delta_test_namespace"))

        # Create table with multiple deltas
        batch1 = pd.DataFrame({"id": [1, 2, 3], "value": ["a", "b", "c"]})

        batch2 = pd.DataFrame({"id": [4, 5, 6], "value": ["d", "e", "f"]})

        # Write first batch (CREATE)
        write_to_table(
            data=batch1,
            table="multi_delta_table",
            namespace="delta_test_namespace",
            mode=TableWriteMode.CREATE,
            content_type=ContentType.PARQUET,
            catalog="test_catalog_1",
        )

        # Write second batch (APPEND - creates second delta)
        write_to_table(
            data=batch2,
            table="multi_delta_table",
            namespace="delta_test_namespace",
            mode=TableWriteMode.APPEND,
            content_type=ContentType.PARQUET,
            catalog="test_catalog_1",
        )

        # Test recursive listing
        all_objects = dc.list(DeltaCatUrl("dc://test_catalog_1"), recursive=True)

        # Extract table names from results
        object_types_to_names = defaultdict(list)
        for obj in all_objects:
            obj_type = Metafile.get_class(obj)
            object_types_to_names[obj_type].append(obj.name)

        # Assert we found all deltas
        expected_deltas = {
            "1",
            "2",
        }  # all deltas should have the same stream position in their respective partitions
        assert (
            len(object_types_to_names[Delta]) == 2
        ), f"Expected 2 deltas, found {len(object_types_to_names[Delta])}"
        assert (
            set(object_types_to_names[Delta]) == expected_deltas
        ), f"Expected deltas: {expected_deltas}, found: {object_types_to_names[Delta]}"

    def test_recursive_listing_empty_namespaces_mixed_with_populated(self):
        """
        Test that recursive listing handles a mix of empty and populated namespaces correctly.
        """
        # Create mix of empty and populated namespaces
        dc.put(DeltaCatUrl("dc://test_catalog_1/empty_namespace_1"))
        dc.put(DeltaCatUrl("dc://test_catalog_1/empty_namespace_2"))
        dc.put(DeltaCatUrl("dc://test_catalog_1/populated_namespace"))

        # Add data only to the populated namespace
        test_data = pd.DataFrame({"id": [1, 2], "data": ["test1", "test2"]})

        write_to_table(
            data=test_data,
            table="test_table",
            namespace="populated_namespace",
            mode=TableWriteMode.CREATE,
            content_type=ContentType.PARQUET,
            catalog="test_catalog_1",
        )

        # Test recursive listing
        all_objects = dc.list(DeltaCatUrl("dc://test_catalog_1"), recursive=True)

        object_types_to_names = defaultdict(list)
        for obj in all_objects:
            obj_type = Metafile.get_class(obj)
            object_types_to_names[obj_type].append(obj.name)

        # Verify we found all namespaces
        expected_namespaces = {
            "empty_namespace_1",
            "empty_namespace_2",
            "populated_namespace",
        }
        assert (
            len(object_types_to_names[Namespace]) == 3
        ), f"Expected 3 namespaces, found {len(object_types_to_names[Namespace])}"
        assert (
            set(object_types_to_names[Namespace]) == expected_namespaces
        ), f"Expected namespaces: {expected_namespaces}, found: {object_types_to_names[Namespace]}"

        # Verify we found the table in the populated namespace
        expected_tables = {"test_table"}
        assert (
            len(object_types_to_names[Table]) == 1
        ), f"Expected 1 table, found {len(object_types_to_names[Table])}"
        assert (
            set(object_types_to_names[Table]) == expected_tables
        ), f"Expected tables: {expected_tables}, found: {object_types_to_names[Table]}"

    def test_non_recursive_listing_vs_recursive_listing(self):
        """
        Test that non-recursive listing only returns top-level objects while recursive returns all.
        """
        # Create nested structure
        dc.put(DeltaCatUrl("dc://test_catalog_1/namespace_one"))
        dc.put(DeltaCatUrl("dc://test_catalog_1/namespace_two"))

        test_data = pd.DataFrame({"id": [1], "value": ["test"]})

        write_to_table(
            data=test_data,
            table="table_in_ns1",
            namespace="namespace_one",
            mode=TableWriteMode.CREATE,
            content_type=ContentType.PARQUET,
            catalog="test_catalog_1",
        )

        # Non-recursive listing (should only get namespaces)
        shallow_objects = dc.list(DeltaCatUrl("dc://test_catalog_1"), recursive=False)

        # Recursive listing (should get everything)
        deep_objects = dc.list(DeltaCatUrl("dc://test_catalog_1"), recursive=True)

        # Shallow should have fewer objects than deep
        assert len(shallow_objects) < len(deep_objects)

        # Shallow should only contain namespaces
        shallow_object_types_to_names = defaultdict(list)
        for obj in shallow_objects:
            obj_type = Metafile.get_class(obj)
            shallow_object_types_to_names[obj_type].append(obj.name)

        # Assert we found all namespaces
        expected_namespaces = {"namespace_one", "namespace_two"}
        assert (
            len(shallow_object_types_to_names[Namespace]) == 2
        ), f"Expected 2 namespaces, found {len(shallow_object_types_to_names[Namespace])}"
        assert (
            set(shallow_object_types_to_names[Namespace]) == expected_namespaces
        ), f"Expected namespaces: {expected_namespaces}, found: {shallow_object_types_to_names[Namespace]}"
        assert (
            len(shallow_object_types_to_names) == 1
        ), f"Expected 1 object type, found {len(shallow_object_types_to_names)}"
        assert set(shallow_object_types_to_names.keys()) == {
            Namespace
        }, f"Expected only Namespace object type, found: {shallow_object_types_to_names.keys()}"

        # Deep should contain multiple types (namespaces, tables, streams, partitions, deltas)
        deep_object_types_to_names = defaultdict(list)
        for obj in deep_objects:
            deep_object_types_to_names[Metafile.get_class(obj)].append(obj.name)

        expected_namespaces = {"namespace_one", "namespace_two"}
        assert (
            len(deep_object_types_to_names[Namespace]) == 2
        ), f"Expected 2 namespaces, found {len(deep_object_types_to_names[Namespace])}"
        assert (
            set(deep_object_types_to_names[Namespace]) == expected_namespaces
        ), f"Expected namespaces: {expected_namespaces}, found: {deep_object_types_to_names[Namespace]}"

        expected_tables = {"table_in_ns1"}
        assert (
            len(deep_object_types_to_names[Table]) == 1
        ), f"Expected 1 table, found {len(deep_object_types_to_names[Table])}"
        assert (
            set(deep_object_types_to_names[Table]) == expected_tables
        ), f"Expected tables: {expected_tables}, found: {deep_object_types_to_names[Table]}"

        expected_table_versions = {"1"}
        assert (
            len(deep_object_types_to_names[TableVersion]) == 1
        ), f"Expected 1 table version, found {len(deep_object_types_to_names[TableVersion])}"
        assert (
            set(deep_object_types_to_names[TableVersion]) == expected_table_versions
        ), f"Expected table versions: {expected_table_versions}, found: {deep_object_types_to_names[TableVersion]}"

        expected_streams = {"deltacat"}
        assert (
            len(deep_object_types_to_names[Stream]) == 1
        ), f"Expected 1 stream, found {len(deep_object_types_to_names[Stream])}"
        assert (
            set(deep_object_types_to_names[Stream]) == expected_streams
        ), f"Expected streams: {expected_streams}, found: {deep_object_types_to_names[Stream]}"

        expected_partitions = {f"None|{UNPARTITIONED_SCHEME_ID}"}
        assert (
            len(deep_object_types_to_names[Partition]) == 1
        ), f"Expected 1 partition, found {len(deep_object_types_to_names[Partition])}"
        assert (
            set(deep_object_types_to_names[Partition]) == expected_partitions
        ), f"Expected partitions: {expected_partitions}, found: {deep_object_types_to_names[Partition]}"

        expected_deltas = {"1"}
        assert (
            len(deep_object_types_to_names[Delta]) == 1
        ), f"Expected 1 delta, found {len(deep_object_types_to_names[Delta])}"
        assert (
            set(deep_object_types_to_names[Delta]) == expected_deltas
        ), f"Expected deltas: {expected_deltas}, found: {deep_object_types_to_names[Delta]}"

    def test_recursive_listing_all_children_processed(self):
        """
        Ensure that all children are processed at each level of recursive listings.
        """
        # Create 3 namespaces
        dc.put(DeltaCatUrl("dc://test_catalog_1/alpha_namespace"))
        dc.put(DeltaCatUrl("dc://test_catalog_1/beta_namespace"))
        dc.put(DeltaCatUrl("dc://test_catalog_1/gamma_namespace"))

        # Create test data
        test_data = pd.DataFrame(
            {"id": [1, 2], "name": ["test1", "test2"], "value": [100, 200]}
        )

        # Create tables in EACH namespace
        write_to_table(
            data=test_data,
            table="alpha_table",
            namespace="alpha_namespace",
            mode=TableWriteMode.CREATE,
            content_type=ContentType.PARQUET,
            catalog="test_catalog_1",
        )

        write_to_table(
            data=test_data,
            table="beta_table",
            namespace="beta_namespace",
            mode=TableWriteMode.CREATE,
            content_type=ContentType.PARQUET,
            catalog="test_catalog_1",
        )

        write_to_table(
            data=test_data,
            table="gamma_table",
            namespace="gamma_namespace",
            mode=TableWriteMode.CREATE,
            content_type=ContentType.PARQUET,
            catalog="test_catalog_1",
        )

        # Perform recursive listing
        all_objects = dc.list(DeltaCatUrl("dc://test_catalog_1"), recursive=True)

        # Extract all objects found
        object_types_to_names = defaultdict(list)
        for obj in all_objects:
            obj_type = Metafile.get_class(obj)
            object_types_to_names[obj_type].append(obj.name)

        # All namespaces should be found
        expected_namespaces = {"alpha_namespace", "beta_namespace", "gamma_namespace"}
        assert (
            len(object_types_to_names[Namespace]) == 3
        ), f"Expected 3 namespaces, found {len(object_types_to_names[Namespace])}"
        assert (
            set(object_types_to_names[Namespace]) == expected_namespaces
        ), f"Expected namespaces: {expected_namespaces}, found: {object_types_to_names[Namespace]}"

        # All tables should be found
        expected_tables = {"alpha_table", "beta_table", "gamma_table"}
        assert (
            len(object_types_to_names[Table]) == 3
        ), f"Expected 3 tables, found {len(object_types_to_names[Table])}"
        assert (
            set(object_types_to_names[Table]) == expected_tables
        ), f"Expected tables: {expected_tables}, found: {object_types_to_names[Table]}"

        # All table versions should be found
        expected_table_versions = {"1"}
        assert (
            len(object_types_to_names[TableVersion]) == 3
        ), f"Expected 3 table versions, found {len(object_types_to_names[TableVersion])}"
        assert (
            set(object_types_to_names[TableVersion]) == expected_table_versions
        ), f"Expected table versions: {expected_table_versions}, found: {object_types_to_names[TableVersion]}"

        # All streams should be found
        expected_streams = {"deltacat"}
        assert (
            len(object_types_to_names[Stream]) == 3
        ), f"Expected 1 stream, found {len(object_types_to_names[Stream])}"
        assert (
            set(object_types_to_names[Stream]) == expected_streams
        ), f"Expected streams: {expected_streams}, found: {object_types_to_names[Stream]}"

        # All partitions should be found
        expected_partitions = {f"None|{UNPARTITIONED_SCHEME_ID}"}
        assert (
            len(object_types_to_names[Partition]) == 3
        ), f"Expected 1 partition, found {len(object_types_to_names[Partition])}"
        assert (
            set(object_types_to_names[Partition]) == expected_partitions
        ), f"Expected partitions: {expected_partitions}, found: {object_types_to_names[Partition]}"

        # All deltas should be found
        expected_deltas = {"1"}
        assert (
            len(object_types_to_names[Delta]) == 3
        ), f"Expected 3 deltas, found {len(object_types_to_names[Delta])}"
        assert (
            set(object_types_to_names[Delta]) == expected_deltas
        ), f"Expected deltas: {expected_deltas}, found: {object_types_to_names[Delta]}"

        # Ensure we found the expected objects across all levels of hierarchy
        total_objects = len(all_objects)
        assert (
            total_objects == 18
        ), f"Expected 18 objects from deep traversal, found only {total_objects}."

    def test_recursive_cross_catalog_copy(self):
        """
        Test comprehensive cross-catalog copy using dc.copy with ** pattern.
        This test validates complete catalog copying with all metadata types:
        namespaces, tables, table versions, streams, partitions, and deltas.
        """
        # Create multiple namespaces, multiple tables, versions, streams, partitions, and deltas

        # Namespace 1: Analytics data with multiple table versions
        dc.put(DeltaCatUrl("dc://test_catalog_1/analytics"))

        # Create table with multiple versions
        events_data_v1 = pd.DataFrame(
            {
                "event_id": [1, 2, 3],
                "user_id": ["user_1", "user_2", "user_3"],
                "event_type": ["click", "view", "purchase"],
                "timestamp": pd.to_datetime(["2023-01-01", "2023-01-02", "2023-01-03"]),
                "value": [10.5, 20.0, 150.75],
            }
        )

        table_properties: TableProperties = {
            TableProperty.READ_OPTIMIZATION_LEVEL: TableReadOptimizationLevel.MAX,
            TableProperty.APPENDED_RECORD_COUNT_COMPACTION_TRIGGER: 1,
            TableProperty.APPENDED_FILE_COUNT_COMPACTION_TRIGGER: 1,
            TableProperty.APPENDED_DELTA_COUNT_COMPACTION_TRIGGER: 1,
        }

        write_to_table(
            data=events_data_v1,
            table="events",
            namespace="analytics",
            mode=TableWriteMode.CREATE,
            content_type=ContentType.PARQUET,
            catalog="test_catalog_1",
            table_properties=table_properties,
        )

        # Add more data to create additional deltas
        events_data_v2 = pd.DataFrame(
            {
                "event_id": [4, 5, 6, 7],
                "user_id": ["user_4", "user_1", "user_5", "user_2"],
                "event_type": ["view", "click", "purchase", "refund"],
                "timestamp": pd.to_datetime(
                    ["2023-01-04", "2023-01-05", "2023-01-06", "2023-01-07"]
                ),
                "value": [0.0, 5.25, 299.99, -150.75],
            }
        )

        write_to_table(
            data=events_data_v2,
            table="events",
            namespace="analytics",
            mode=TableWriteMode.APPEND,
            content_type=ContentType.PARQUET,
            catalog="test_catalog_1",
        )

        # Create second table in analytics namespace
        users_data = pd.DataFrame(
            {
                "user_id": ["user_1", "user_2", "user_3", "user_4", "user_5"],
                "username": ["alice", "bob", "charlie", "diana", "eve"],
                "email": [
                    "alice@test.com",
                    "bob@test.com",
                    "charlie@test.com",
                    "diana@test.com",
                    "eve@test.com",
                ],
                "created_at": pd.to_datetime(
                    [
                        "2022-12-01",
                        "2022-12-15",
                        "2022-12-20",
                        "2023-01-01",
                        "2023-01-03",
                    ]
                ),
                "is_active": [True, True, False, True, True],
            }
        )

        write_to_table(
            data=users_data,
            table="users",
            namespace="analytics",
            mode=TableWriteMode.CREATE,
            content_type=ContentType.PARQUET,
            catalog="test_catalog_1",
        )

        # Create version 2 of the events table to test table version ordering in recursive copy
        events_data_v3 = pd.DataFrame(
            {
                "event_id": [8, 9, 10],
                "user_id": ["user_3", "user_4", "user_5"],
                "event_type": ["signup", "login", "logout"],
                "timestamp": pd.to_datetime(["2023-01-08", "2023-01-09", "2023-01-10"]),
                "value": [0.0, 0.0, 0.0],
            }
        )

        write_to_table(
            data=events_data_v3,
            table="events",
            namespace="analytics",
            table_version="2",  # Explicitly create version 2
            mode=TableWriteMode.CREATE,
            content_type=ContentType.PARQUET,
            catalog="test_catalog_1",
        )

        # Namespace 2: Product data with different schema
        dc.put(DeltaCatUrl("dc://test_catalog_1/products"))

        products_data = pd.DataFrame(
            {
                "product_id": ["prod_1", "prod_2", "prod_3"],
                "name": ["Widget A", "Widget B", "Super Widget"],
                "category": ["widgets", "widgets", "premium"],
                "price": [19.99, 29.99, 149.99],
                "in_stock": [True, False, True],
                "metadata": [
                    {"color": "red"},
                    {"color": "blue", "size": "large"},
                    {"color": "gold", "premium": True},
                ],
            }
        )

        write_to_table(
            data=products_data,
            table="inventory",
            namespace="products",
            mode=TableWriteMode.CREATE,
            content_type=ContentType.PARQUET,
            catalog="test_catalog_1",
        )

        # Create product categories table
        categories_data = pd.DataFrame(
            {
                "category_id": ["widgets", "premium", "accessories"],
                "display_name": ["Standard Widgets", "Premium Products", "Accessories"],
                "description": [
                    "Basic widget products",
                    "High-end premium items",
                    "Additional accessories",
                ],
                "active": [True, True, False],
            }
        )

        write_to_table(
            data=categories_data,
            table="categories",
            namespace="products",
            mode=TableWriteMode.CREATE,
            content_type=ContentType.PARQUET,
            catalog="test_catalog_1",
        )

        # Namespace 3: Empty namespace (edge case testing)
        dc.put(DeltaCatUrl("dc://test_catalog_1/empty_data"))

        # Namespace 4: Orders with complex nested data
        dc.put(DeltaCatUrl("dc://test_catalog_1/orders"))

        orders_data = pd.DataFrame(
            {
                "order_id": ["order_1", "order_2", "order_3"],
                "user_id": ["user_1", "user_2", "user_1"],
                "product_ids": [["prod_1"], ["prod_2", "prod_3"], ["prod_1", "prod_2"]],
                "order_date": pd.to_datetime(
                    ["2023-01-05", "2023-01-06", "2023-01-07"]
                ),
                "total_amount": [19.99, 179.98, 49.98],
                "status": ["completed", "pending", "completed"],
            }
        )

        write_to_table(
            data=orders_data,
            table="transactions",
            namespace="orders",
            mode=TableWriteMode.CREATE,
            content_type=ContentType.PARQUET,
            catalog="test_catalog_1",
        )

        # Verify source catalog structure before copy
        source_objects = dc.list(DeltaCatUrl("dc://test_catalog_1"), recursive=True)
        source_urls_by_type = defaultdict(list)
        source_by_type = defaultdict(list)

        for obj in source_objects:
            obj_class = Metafile.get_class(obj.to_serializable())
            source_urls_by_type[obj_class].append(obj.url())
            source_by_type[obj_class].append(obj)

        assert (
            len(source_urls_by_type[Namespace]) == 4
        ), f"Expected 4 namespaces, got {len(source_urls_by_type[Namespace])}"
        assert (
            len(source_urls_by_type[Table]) == 5
        ), f"Expected 5 tables, got {len(source_urls_by_type[Table])}"
        assert (
            len(source_urls_by_type[TableVersion]) == 6
        ), f"Expected 6 table versions, got {len(source_urls_by_type[TableVersion])}"
        assert (
            len(source_urls_by_type[Stream]) == 6
        ), f"Expected 6 streams, got {len(source_urls_by_type[Stream])}"
        assert (
            len(source_urls_by_type[Partition]) == 6
        ), f"Expected 6 partitions, got {len(source_urls_by_type[Partition])}"
        assert (
            len(source_urls_by_type[Delta]) == 6
        ), f"Expected 6 deltas, got {len(source_urls_by_type[Delta])}"

        # Test the /** recursive copy pattern.
        dc.copy(
            DeltaCatUrl("dc://test_catalog_1/**"),  # ** means recursive copy all
            DeltaCatUrl("dc://test_catalog_2/"),
        )

        # Verify destination catalog has same structure
        dest_objects = dc.list(DeltaCatUrl("dc://test_catalog_2"), recursive=True)
        dest_urls_by_type = defaultdict(list)
        dest_by_type = defaultdict(list)

        assert len(dest_objects) == len(
            source_objects
        ), f"Expected {len(source_objects)} objects, got {len(dest_objects)}"

        for obj in dest_objects:
            obj_class = Metafile.get_class(obj.to_serializable())
            dest_urls_by_type[obj_class].append(obj.url())
            dest_by_type[obj_class].append(obj)

        assert sorted(dest_urls_by_type[Namespace]) == sorted(
            source_urls_by_type[Namespace]
        ), f"Namespace mismatch: {dest_urls_by_type[Namespace]} vs {source_urls_by_type[Namespace]}"
        assert sorted(dest_urls_by_type[Table]) == sorted(
            source_urls_by_type[Table]
        ), f"Table mismatch: {dest_urls_by_type[Table]} vs {source_urls_by_type[Table]}"
        assert sorted(dest_urls_by_type[TableVersion]) == sorted(
            source_urls_by_type[TableVersion]
        ), f"Table version mismatch: {dest_urls_by_type[TableVersion]} vs {source_urls_by_type[TableVersion]}"
        assert sorted(dest_urls_by_type[Stream]) == sorted(
            source_urls_by_type[Stream]
        ), f"Stream mismatch: {dest_urls_by_type[Stream]} vs {source_urls_by_type[Stream]}"
        assert sorted(dest_urls_by_type[Partition]) == sorted(
            source_urls_by_type[Partition]
        ), f"Partition mismatch: {dest_urls_by_type[Partition]} vs {source_urls_by_type[Partition]}"
        assert sorted(dest_urls_by_type[Delta]) == sorted(
            source_urls_by_type[Delta]
        ), f"Delta mismatch: {dest_urls_by_type[Delta]} vs {source_urls_by_type[Delta]}"

        # Validate each hierarchy level
        for obj_type in source_by_type.keys():
            source_count = len(source_by_type.get(obj_type))
            dest_count = len(dest_by_type.get(obj_type, []))
            assert (
                dest_count == source_count
            ), f"{obj_type} count mismatch: {dest_count} vs {source_count}"

            # Spot check equivalence of each type
            if obj_type == Namespace and source_count > 0:
                # Check namespace properties are preserved
                source_ns = source_by_type[obj_type][0]  # NamespaceModel
                dest_ns = next(
                    (
                        ns
                        for ns in dest_by_type[obj_type]
                        if ns.namespace == source_ns.namespace
                    ),
                    None,
                )
                assert (
                    dest_ns is not None
                ), f"Namespace {source_ns.namespace} not found in destination"
                assert source_ns.equivalent_to(
                    dest_ns
                ), f"Namespace {source_ns.namespace} not equivalent to {dest_ns.namespace}"
            elif obj_type == Table:
                source_table = source_by_type[obj_type][0]  # TableModel
                dest_table = next(
                    (
                        t
                        for t in dest_by_type[obj_type]
                        if t.namespace == source_table.namespace
                        and t.table_name == source_table.table_name
                    ),
                    None,
                )
                assert (
                    dest_table is not None
                ), f"Table {source_table.namespace}.{source_table.table_name} not found in destination"
                assert source_table.equivalent_to(
                    dest_table
                ), f"Table {source_table.namespace}.{source_table.table_name} not equivalent to {dest_table.namespace}.{dest_table.table_name}"
            elif obj_type == TableVersion and source_count > 0:
                # Check table version properties are preserved
                source_tv = source_by_type[obj_type][0]  # TableVersionModel
                dest_tv = next(
                    (
                        tv
                        for tv in dest_by_type[obj_type]
                        if tv.namespace == source_tv.namespace
                        and tv.table_name == source_tv.table_name
                        and tv.table_version == source_tv.table_version
                    ),
                    None,
                )
                assert (
                    dest_tv is not None
                ), f"TableVersion {source_tv.namespace}.{source_tv.table_name}.{source_tv.table_version} not found in destination"
                assert dest_tv.equivalent_to(
                    source_tv
                ), f"TableVersion {source_tv.namespace}.{source_tv.table_name}.{source_tv.table_version} not equivalent to {dest_tv.namespace}.{dest_tv.table_name}.{dest_tv.table_version}"

                # Special validation for table version ordering - check that analytics.events has versions 1 and 2
                analytics_events_versions = [
                    tv
                    for tv in dest_by_type[obj_type]
                    if tv.namespace == "analytics" and tv.table_name == "events"
                ]
                if analytics_events_versions:
                    versions = sorted(
                        [tv.table_version for tv in analytics_events_versions]
                    )
                    assert versions == [
                        "1",
                        "2",
                    ], f"Expected analytics.events versions ['1', '2'], got {versions}"
            elif obj_type == Stream and source_count > 0:
                # Check stream properties are preserved
                source_stream = source_by_type[obj_type][0]  # StreamModel
                dest_stream = next(
                    (
                        s
                        for s in dest_by_type[obj_type]
                        if s.namespace == source_stream.namespace
                        and s.table_name == source_stream.table_name
                        and s.stream_format == source_stream.stream_format
                    ),
                    None,
                )
                assert (
                    dest_stream is not None
                ), f"Stream {source_stream.namespace}.{source_stream.table_name}.{source_stream.stream_format} not found in destination"
                assert dest_stream.equivalent_to(
                    source_stream
                ), f"Stream {source_stream.namespace}.{source_stream.table_name}.{source_stream.stream_format} not equivalent to {dest_stream.namespace}.{dest_stream.table_name}.{dest_stream.stream_format}"
            elif obj_type == Partition and source_count > 0:
                # Check partition properties are preserved (with new partition IDs)
                source_partition = source_by_type[obj_type][0]  # PartitionModel
                dest_partition = next(
                    (
                        p
                        for p in dest_by_type[obj_type]
                        if p.namespace == source_partition.namespace
                        and p.table_name == source_partition.table_name
                    ),
                    None,
                )
                assert (
                    dest_partition is not None
                ), f"Partition for {source_partition.namespace}.{source_partition.table_name} not found in destination"
                assert dest_partition.equivalent_to(
                    source_partition
                ), f"Partition {source_partition.namespace}.{source_partition.table_name} not equivalent to {dest_partition.namespace}.{dest_partition.table_name}"
            elif obj_type == Delta and source_count > 0:
                # Check delta properties are preserved (with same stream positions)
                source_delta = source_by_type[obj_type][0]  # DeltaModel
                dest_delta = next(
                    (
                        d
                        for d in dest_by_type[obj_type]
                        if d.namespace == source_delta.namespace
                        and d.table_name == source_delta.table_name
                        and d.stream_position == source_delta.stream_position
                    ),
                    None,
                )
                assert (
                    dest_delta is not None
                ), f"Delta for {source_delta.namespace}.{source_delta.table_name} at position {source_delta.stream_position} not found in destination"
                assert dest_delta.equivalent_to(
                    source_delta
                ), f"Delta {source_delta.namespace}.{source_delta.table_name} at position {source_delta.stream_position} not equivalent to {dest_delta.namespace}.{dest_delta.table_name} at position {dest_delta.stream_position}"

        # Validate each table's data integrity
        test_cases = [
            ("analytics", "events"),
            ("analytics", "users"),
            ("products", "inventory"),
            ("products", "categories"),
            ("orders", "transactions"),
        ]

        for namespace, table in test_cases:
            # Check table exists in destination
            assert dc.table_exists(
                table=table,
                namespace=namespace,
                catalog="test_catalog_2",
            ), f"Table {namespace}/{table} should exist in destination catalog"

            # Verify table data equivalence using read_table
            source_df = dc.read_table(
                table=table,
                namespace=namespace,
                catalog="test_catalog_1",
                read_as=DatasetType.PANDAS,
            )

            dest_df = dc.read_table(
                table=table,
                namespace=namespace,
                catalog="test_catalog_2",
                read_as=DatasetType.PANDAS,
            )

            # Verify both datasets are valid pandas DataFrames
            assert (
                source_df is not None
            ), f"Source data should not be None for {namespace}.{table}"
            assert (
                dest_df is not None
            ), f"Destination data should not be None for {namespace}.{table}"

            # Compare DataFrame properties
            assert len(source_df) == len(
                dest_df
            ), f"Row count mismatch for {namespace}.{table}: {len(source_df)} vs {len(dest_df)}"
            assert list(source_df.columns) == list(
                dest_df.columns
            ), f"Column mismatch for {namespace}.{table}"

            # Sort both dataframes by first column for comparison (to handle potential row ordering differences)
            _assert_data_equivalence(source_df, dest_df)

            # Verify that writing to the source table doesn't affect the destination table
            dc.write_to_table(
                data=source_df,
                table=table,
                namespace=namespace,
                catalog="test_catalog_1",
                mode=TableWriteMode.APPEND,
            )

            # Verify that the destination table's data hasn't changed
            dest_df = dc.read_table(
                table=table,
                namespace=namespace,
                catalog="test_catalog_2",
                read_as=DatasetType.PANDAS,
            )
            _assert_data_equivalence(source_df, dest_df)

            # Verify that the source table has source_df repeated twice
            source_df_repeated = dc.read_table(
                table=table,
                namespace=namespace,
                catalog="test_catalog_1",
                read_as=DatasetType.PANDAS,
            )
            assert (
                len(source_df_repeated) == len(source_df) * 2
            ), f"Source table {namespace}.{table} should have {len(source_df) * 2} rows"

            # Verify that writing to the destination table doesn't affect the source table
            dc.write_to_table(
                data=dest_df,
                table=table,
                namespace=namespace,
                catalog="test_catalog_2",
                mode=TableWriteMode.APPEND,
            )

            # Verify that the source table's data hasn't changed
            source_df_unchanged = dc.read_table(
                table=table,
                namespace=namespace,
                catalog="test_catalog_1",
                read_as=DatasetType.PANDAS,
            )
            _assert_data_equivalence(source_df_repeated, source_df_unchanged)

            # Verify that the destination table's data has dest_df repeated twice
            dest_df_repeated = dc.read_table(
                table=table,
                namespace=namespace,
                catalog="test_catalog_2",
                read_as=DatasetType.PANDAS,
            )
            assert (
                len(dest_df_repeated) == len(dest_df) * 2
            ), f"Destination table {namespace}.{table} should have {len(dest_df) * 2} rows"

        # Verify empty namespace was copied correctly
        assert dc.namespace_exists(
            namespace="empty_data",
            catalog="test_catalog_2",
        ), "Empty namespace should exist in destination catalog"


def _assert_data_equivalence(source_df: pd.DataFrame, dest_df: pd.DataFrame):
    # Sort both dataframes by first column for comparison (to handle potential row ordering differences)
    if len(source_df) > 0:
        first_col = source_df.columns[0]
        # Handle sorting with potential complex data types
        source_sorted = source_df.sort_values(first_col).reset_index(drop=True)
        dest_sorted = dest_df.sort_values(first_col).reset_index(drop=True)

        # Compare data values using pandas testing
        pd.testing.assert_frame_equal(
            source_sorted,
            dest_sorted,
        )



from abc import abstractmethod
from typing import Iterable, Optional, Protocol, TypeVar, Union

from deltacat.experimental.storage.rivulet.reader.dataset_metastore import (
    DatasetMetastore,
)

# TODO: Add type validation in dataset/schema classes
T = TypeVar("T", bound=Union[int, str])


class Shard(Protocol[T]):
    """
    Abstract base class representing a shard with defined inclusive boundaries.

    A shard represents a logical partition of data, defined by its
    minimum and maximum keys. These keys determine the range of data
    within a dataset that the shard encompasses.
    """

    min_key: Optional[T]
    max_key: Optional[T]


class ShardingStrategy(Protocol):
    """
    A sharding strategy determines how the dataset is divided into shards.
    """

    @staticmethod
    def from_string(strategy: str) -> "ShardingStrategy":
        """
        Factory method to create the appropriate ShardingStrategy from a string.

        param: strategy: The string representation of the sharding strategy.
        return: ShardingStrategy class.
        """
        if strategy == "range":
            from deltacat.experimental.storage.rivulet.shard.range_shard import (
                RangeShardingStrategy,
            )

            return RangeShardingStrategy()
        else:
            raise ValueError(f"Unsupported sharding strategy type: {strategy}")

    @abstractmethod
    def shards(self, num_shards: int, metastore: DatasetMetastore) -> Iterable[Shard]:
        """
        Generate the shards based on the chosen strategy.
        """



from typing import Optional, Dict

import pyarrow as pa

from deltacat import (
    ContentEncoding,
    ContentType,
)
from deltacat.storage import (
    BucketTransform,
    BucketTransformParameters,
    BucketingStrategy,
    CommitState,
    Delta,
    DeltaLocator,
    DeltaType,
    EntryParams,
    EntryType,
    Field,
    LifecycleState,
    ManifestAuthor,
    ManifestEntry,
    Namespace,
    NamespaceLocator,
    NullOrder,
    Partition,
    PartitionKey,
    PartitionKeyList,
    PartitionLocator,
    PartitionScheme,
    PartitionSchemeList,
    Schema,
    SchemaList,
    SortScheme,
    SortSchemeList,
    SortKey,
    SortKeyList,
    SortOrder,
    StreamLocator,
    StreamFormat,
    Stream,
    Table,
    TableLocator,
    TableVersionLocator,
    TableVersion,
    TruncateTransform,
    TruncateTransformParameters,
)

from deltacat.storage.model.manifest import (
    Manifest,
    ManifestMeta,
    ManifestEntryList,
)
from deltacat.utils.common import current_time_ms


def create_empty_delta(
    partition: Partition,
    delta_type: DeltaType,
    author: Optional[str],
    properties: Optional[Dict[str, str]] = None,
    manifest_entry_id: Optional[str] = None,
) -> Delta:
    stream_position = current_time_ms()
    delta_locator = DeltaLocator.of(
        partition.locator,
        stream_position=stream_position,
    )

    if manifest_entry_id:
        manifest = Manifest.of(
            entries=ManifestEntryList.of([]),
            author=author,
            uuid=manifest_entry_id,
        )
    else:
        manifest = None

    return Delta.of(
        delta_locator,
        delta_type=delta_type,
        meta=ManifestMeta(),
        properties=properties,
        manifest=manifest,
        previous_stream_position=partition.stream_position,
    )


def create_test_namespace():
    namespace_locator = NamespaceLocator.of(namespace="test_namespace")
    return Namespace.of(locator=namespace_locator)


def create_test_table():
    table_locator = TableLocator.at(
        namespace="test_namespace",
        table_name="test_table",
    )
    return Table.of(
        locator=table_locator,
        description="test table description",
    )


def create_test_table_version():
    table_version_locator = TableVersionLocator.at(
        namespace="test_namespace",
        table_name="test_table",
        table_version="v.1",
    )
    schema = Schema.of(
        [
            Field.of(
                field=pa.field("some_string", pa.string(), nullable=False),
                field_id=1,
                is_merge_key=True,
            ),
            Field.of(
                field=pa.field("some_int32", pa.int32(), nullable=False),
                field_id=2,
                is_merge_key=True,
            ),
            Field.of(
                field=pa.field("some_float64", pa.float64()),
                field_id=3,
                is_merge_key=False,
            ),
        ]
    )
    bucket_transform = BucketTransform.of(
        BucketTransformParameters.of(
            num_buckets=2,
            bucketing_strategy=BucketingStrategy.DEFAULT,
        )
    )
    partition_keys = [
        PartitionKey.of(
            key=["some_string", "some_int32"],
            name="test_partition_key",
            field_id=1,
            transform=bucket_transform,
        )
    ]
    partition_scheme = PartitionScheme.of(
        keys=PartitionKeyList.of(partition_keys),
        name="test_partition_scheme",
        scheme_id="test_partition_scheme_id",
    )
    sort_keys = [
        SortKey.of(
            key=["some_int32"],
            sort_order=SortOrder.DESCENDING,
            null_order=NullOrder.AT_START,
            transform=TruncateTransform.of(
                TruncateTransformParameters.of(width=3),
            ),
        )
    ]
    sort_scheme = SortScheme.of(
        keys=SortKeyList.of(sort_keys),
        name="test_sort_scheme",
        scheme_id="test_sort_scheme_id",
    )
    return TableVersion.of(
        locator=table_version_locator,
        schema=schema,
        partition_scheme=partition_scheme,
        description="test table version description",
        properties={"test_property_key": "test_property_value"},
        content_types=[ContentType.PARQUET],
        sort_scheme=sort_scheme,
        watermark=None,
        lifecycle_state=LifecycleState.CREATED,
        schemas=SchemaList.of([schema]),
        partition_schemes=PartitionSchemeList.of([partition_scheme]),
        sort_schemes=SortSchemeList.of([sort_scheme]),
    )


def create_test_stream():
    stream_locator = StreamLocator.at(
        namespace="test_namespace",
        table_name="test_table",
        table_version="v.1",
        stream_id="test_stream_id",
        stream_format=StreamFormat.DELTACAT,
    )
    bucket_transform = BucketTransform.of(
        BucketTransformParameters.of(
            num_buckets=2,
            bucketing_strategy=BucketingStrategy.DEFAULT,
        )
    )
    partition_keys = [
        PartitionKey.of(
            key=["some_string", "some_int32"],
            name="test_partition_key",
            field_id=1,
            transform=bucket_transform,
        )
    ]
    partition_scheme = PartitionScheme.of(
        keys=PartitionKeyList.of(partition_keys),
        name="test_partition_scheme",
        scheme_id="test_partition_scheme_id",
    )
    return Stream.of(
        locator=stream_locator,
        partition_scheme=partition_scheme,
        state=CommitState.STAGED,
        previous_stream_id="test_previous_stream_id",
        watermark=1,
    )


def create_test_partition():
    partition_locator = PartitionLocator.at(
        namespace="test_namespace",
        table_name="test_table",
        table_version="v.1",
        stream_id="test_stream_id",
        stream_format=StreamFormat.DELTACAT,
        partition_values=["a", 1],
        partition_id="test_partition_id",
    )
    return Partition.of(
        locator=partition_locator,
        content_types=[ContentType.PARQUET],
        state=CommitState.STAGED,
        previous_stream_position=0,
        previous_partition_id="test_previous_partition_id",
        stream_position=1,
        partition_scheme_id="test_partition_scheme_id",
    )


def create_test_delta():
    delta_locator = DeltaLocator.at(
        namespace="test_namespace",
        table_name="test_table",
        table_version="v.1",
        stream_id="test_stream_id",
        stream_format=StreamFormat.DELTACAT,
        partition_values=["a", 1],
        partition_id="test_partition_id",
        stream_position=1,
    )
    manifest_entry_params = EntryParams.of(
        equality_field_locators=["some_string", "some_int32"],
    )
    manifest_meta = ManifestMeta.of(
        record_count=1,
        content_length=10,
        content_type=ContentType.PARQUET.value,
        content_encoding=ContentEncoding.IDENTITY.value,
        source_content_length=100,
        credentials={"foo": "bar"},
        content_type_parameters=[{"param1": "value1"}],
        entry_type=EntryType.EQUALITY_DELETE,
        entry_params=manifest_entry_params,
    )
    manifest = Manifest.of(
        entries=ManifestEntryList(
            [
                ManifestEntry.of(
                    url="s3://test/url",
                    meta=manifest_meta,
                )
            ]
        ),
        author=ManifestAuthor.of(
            name="deltacat",
            version="2.0",
        ),
        entry_type=EntryType.EQUALITY_DELETE,
        entry_params=manifest_entry_params,
    )
    return Delta.of(
        locator=delta_locator,
        delta_type=DeltaType.APPEND,
        meta=manifest_meta,
        properties={"property1": "value1"},
        manifest=manifest,
        previous_stream_position=0,
    )
